import unittest
import json
import time
import os

import firebase_admin
from firebase_admin import credentials
from firebase_admin import firestore, auth

from fam.exceptions import *
from fam.tests.models.test01 import GenericObject, Dog, Cat, Person, JackRussell, Monkey, Monarch, NAMESPACE

from fam.database import FirestoreWrapper
from fam.mapper import ClassMapper

from.config_local import *


ROOT_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))))


class TestDB(unittest.TestCase):


    def setUp(self):

        creds_path = os.path.join(ROOT_DIR, "secrets", "earth-rover-test-d241bce5266d.json")
        mapper = ClassMapper([Dog, Cat, Person, JackRussell, Monkey, Monarch])
        # self.db = FirestoreWrapper(mapper, creds_path)


        cred = credentials.Certificate(creds_path)
        app = firebase_admin.initialize_app(cred)
        uid = 'some-uid'

        additional_claims = {
            'premiumAccount': True
        }

        custom_token = auth.create_custom_token(uid, additional_claims)
        print(custom_token)
        print(PROJECT_NAME)
        print(API_KEY)
        firebase_admin.delete_app(app)

        self.db = FirestoreWrapper(mapper,
                                   None,
                                   project_id=PROJECT_NAME,
                                   custom_token=custom_token.decode("utf-8"),
                                   api_key=API_KEY,
                                   namespace=NAMESPACE

                                   )


    def test_make_an_object_and_refresh(self):
        dog = Dog.create(self.db, name="woofer")
        dog2 = Dog.get(self.db, dog.key)
        self.assertIsNotNone(dog2)
        time.sleep(1)

        self.db.refresh()
        dog3 = Dog.create(self.db, name="another")
        dog4 = Dog.get(self.db, dog3.key)
        self.assertIsNotNone(dog4)






def object_default(o):
    if hasattr(o, "to_json"):
        return o.to_json()
    raise TypeError(repr(o) + " is not JSON serializable")


# class PatchedJson(object):
#     import simplejson as json
#
#     def dumps(self, *args, **kwargs):
#         print args[0]
#         return self.json.dumps(*args, indent=4, sort_keys=True, default=object_default, **kwargs)
#
#     def __getattr__(self, name):
#         return getattr(self.json, name)
#
#
# sys.modules[__name__] = PatchedJson()









from fam.blud import ReferenceFrom, GenericObject
import json

class FamDbAuthException(Exception):
    pass

class BaseDatabase(object):

    FOREIGN_KEY_MAP_STRING = '''function(doc) {
    var resources = %s;
    if (resources.indexOf(doc.type) != -1 && doc.namespace == \"%s\"){
        emit(doc.%s, doc);
    }
}'''

    check_on_save = True

###################################

    #double dispatch accessors that return objects

    def put(self, thing):
        return thing.save(self)

    def delete(self, thing):
        return thing.delete(self)

    def get(self, key, class_name=None):
        return GenericObject.get(self, key, class_name=class_name)

    def delete_key(self, key):
        return GenericObject.delete_key(self, key)

    def query_view(self, view_name, **kwargs):
        return GenericObject.view(self, view_name, **kwargs)

    def changes(self, since=None, channels=None, limit=None, feed=None, timeout=None, filter=None):
        return GenericObject.changes(self, since=since, channels=channels, limit=limit, feed=feed, timeout=timeout, filter=filter)


#################################

    def class_for_type_name(self, type_name, namespace_name):
        return self.mapper.get_class(type_name, namespace_name)




from fam.database.base import BaseDatabase
from fam.buffer.write_buffer import FamWriteBuffer


class NullDatabase(BaseDatabase):
    database_type = "null"

    def __init__(self, mapper):
        self.mapper = mapper

    def put(self, thing):
        pass

    def delete(self, thing):
        pass

    def get(self, key, class_name=None):
        return None

    def delete_key(self, key):
        pass

    def query_view(self, view_name, **kwargs):
        return []




import os
import unittest
from fam.database import CouchDBWrapper
from fam.mapper import ClassMapper

from fam.tests.test_couchdb.config import *
from fam.tests.models.test01 import Dog, Cat, Person, JackRussell

THIS_DIR = os.path.dirname(__file__)
DATA_PATH = os.path.join(THIS_DIR, "data")


class IndexTests(unittest.TestCase):

    db = None

    def setUp(self):
        filepath = os.path.join(THIS_DIR, "animal_views.js")
        mapper = ClassMapper([Dog, Cat, Person, JackRussell], designs=[filepath])
        self.db = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True)
        self.db.update_designs()

    def tearDown(self):
        pass

    def test_create_index(self):
        filepath = os.path.join(THIS_DIR, "animal_views.js")

        as_dict = self.db.mapper._js_design_as_doc(filepath)

        expected = {
            "_id": "_design/animal_views",
            "views": {
                "cat_legs": {
                    "map": "function(doc) {\n  if (doc.type == \"cat\") {\n    emit(doc.legs, doc);\n  }\n}"
                }
            }
        }

        self.assertEqual(as_dict, expected)


    def test_query_view(self):

        paul = Person(name="Paul")
        self.db.put(paul)
        cat1 = Cat(owner=paul, legs=4)
        self.db.put(cat1)
        cat2 = Cat(owner=paul, legs=3)
        self.db.put(cat2)
        three_legged_cats = Cat.all_with_n_legs(self.db, 3)
        self.assertEqual(len(three_legged_cats), 1)

        self.assertEqual(three_legged_cats[0].key, cat2.key)


    def test_long_polling(self):
        paul = Person(name="Paul")
        self.db.put(paul)
        cat1 = Cat(owner=paul, legs=4)
        self.db.put(cat1)
        cat2 = Cat(owner=paul, legs=3)
        self.db.put(cat2)
        three_legged_cats = self.db.view("animal_views/cat_legs", key=3)
        self.assertEqual(len(three_legged_cats), 1)
        self.assertEqual(three_legged_cats[0].key, cat2.key)


from __future__ import absolute_import
import os
import unittest
from fam.database import CouchDBWrapper
from fam.mapper import ClassMapper
from fam.buffer import buffered_db
from fam.tests.test_couchdb.config import *
from fam.tests.models.test01 import Dog, Cat, Person

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(THIS_DIR, "data")

class CacheTests(unittest.TestCase):

    def setUp(self):
        mapper = ClassMapper([Dog, Cat, Person])
        self.db = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True)
        self.db.update_designs()

    def tearDown(self):
        pass



    def test_iterate_dogs(self):

        me = Person(name="paul")
        self.db.put(me)

        for i in range(500):
            dog = Dog(name="dog_%s" % i, owner=me)
            self.db.put(dog)

        counter = 0
        for dog in me.dogs:
            counter += 1

        self.assertEqual(counter, 500)


import simplejson as json
import copy


from fam.fam_json import object_default
import jsonschema
from copy import deepcopy

from fam.exceptions import *
from fam.constants import *
from fam.utils import requests_shim as requests
from fam.database.base import BaseDatabase, FamDbAuthException

from fam.utils.backoff import http_backoff
from .couchdb_adapter import CouchDBDataAdapter


JSON_KEY_STRINGS = ["endkey", "end_key", "key", "keys", "startkey", "start_key"]


class ResultWrapper(object):

    def __init__(self, key, rev, value):
        self.key = key
        self.rev = rev
        self.value = value


    @classmethod
    def from_couchdb_json(cls, as_json):
        key = as_json["_id"]
        rev = as_json.get("_rev")
        if rev is not None:
            rev = rev
            del as_json["_rev"]
        else:
            rev = None
        del as_json["_id"]
        value = as_json

        # if rev is None:
        #     print "****************************"
        #     print as_json
        #     print "****************************"
        # else:
        #     if rev[1:2] != "-":
        #         print "****************************"
        #         print rev
        #         print as_json
        #         print "****************************"

        return cls(key, rev, value)


    @classmethod
    def from_couchdb_view_json(cls, as_json):
        key = as_json["id"]
        rev = as_json["value"]["_rev"]
        value = deepcopy(as_json["value"])
        del value["_id"]
        del value["_rev"]
        return cls(key, rev, value)


    @classmethod
    def from_gateway_view_json(cls, as_json):
        # the format of this seems to be changing quite a bit
        try:
            key = as_json["id"]
            value = deepcopy(as_json["value"])
            sync = value.get("_sync")
            if sync is not None:
                rev = sync["rev"]
                del value["_sync"]
            elif value.get("_rev"):
                rev = as_json["value"]["_rev"]
                del value["_rev"]
            else:
                rev = None
        except KeyError as e:
            print("key error raised in from_gateway_view_json on object: %s" % json.dumps(as_json, indent=4))
            raise e

        # if rev[1:2] != "-":
        #     print "how is this rev wrong!: ", rev
        #     print as_json


        return cls(key, rev, value)


def auth(func):
    def func_wrapper(instance, *args, **kwargs):
        try:
            return func(instance, *args, **kwargs)
        except FamDbAuthException:
            instance.authenticate()
            return func(instance, *args, **kwargs)
    return func_wrapper


def ensure_views(func):
    def func_wrapper(db, *args, **kwargs):
        try:
            return func(db, *args, **kwargs)
        except FamViewError:
            db.update_designs()
            return func(db, *args, **kwargs)
    return func_wrapper


class CouchDBWrapper(BaseDatabase):

    VIEW_URL = "%s/%s/_design/%s/_view/%s"
    # "%s/%s/_design/%s/_view/%s?stale=false&key=\"%s\""

    database_type = "couchdb"
    supports_skip = True
    check_on_save = True

    def __init__(self, mapper,
                 db_url,
                 db_name,
                 reset=False,
                 remote_url=None,
                 continuous=False,
                 validator=None,
                 read_only=False
                 ):

        self.mapper = mapper
        self.validator = validator
        self.read_only = read_only

        self.cookies = {}

        self.remote_url = remote_url
        self.db_name = db_name
        self.db_url = db_url
        self.session = requests.Session()
        self.data_adapter = CouchDBDataAdapter()

        url = "%s/%s" % (db_url, db_name)

        replicator_url = "{}/_config/replicator/db".format(self.db_url)
        self.replicator_db = self.session.get(replicator_url).json()

        if reset:
            self.replicator_db = self.clear_all_replications()
            rsp = self.session.get(url)
            if rsp.status_code == 200:
                rsp = self.session.delete("%s/%s" % (db_url, db_name))
                if rsp.status_code == 401:
                    raise Exception("Error deleting CB database: 401 Unauthorized")
                if rsp.status_code == 400:
                    raise Exception("Error deleting CB database: 400 Bad Request")

        rsp = self.session.get(url)
        if rsp.status_code == 200:
            print("exists", db_name, db_url)

        if rsp.status_code == 404:
            rsp = self.session.put(url)
            if rsp.status_code == 401:
                raise Exception("Error creating CB database: 401 Unauthorized")
            if rsp.status_code == 400:
                raise Exception("Error creating CB database: 400 Bad Request")
            if not(rsp.status_code == 201 or rsp.status_code == 412):
                raise Exception("Unknown Error creating cb database: %s" % rsp.status_code)

            #if this is a new database then set the replications running
            self.continuous = continuous
            if continuous:
                self.sync_both_continuous()



    def info(self):

        url = "%s/%s" % (self.db_url, self.db_name)
        rsp = self.session.get(url)

        if rsp.status_code == 200:
            return rsp.json()

        return None


    def get_design(self, key):
        return self._get(key)

    def ensure_role(self, role_name):
        pass

    def ensure_user_role(self, username, role):
        pass

    @auth
    def _get(self, key, class_name=None):
        url = "%s/%s/%s" % (self.db_url, self.db_name, key)
        # print "_get: ", url
        rsp = self.session.get(url)
        if rsp.status_code == 200:
            return ResultWrapper.from_couchdb_json(self.data_adapter.deserialise(rsp.json()))
        if rsp.status_code == 404:
            # print "not found: ", key
            return None
        if rsp.status_code == 401:
            raise FamDbAuthException(" %s %s" % (rsp.status_code, rsp.text))
        raise Exception("Unknown Error getting cb doc: %s %s" % (rsp.status_code, rsp.text))


    def get_refs_from(self, namespace, type_name, name, key, field):
        view_namespace = namespace.replace("/", "_")
        view_name = "%s/%s_%s" % (view_namespace, type_name, name)
        return self.query_view(view_name, key=key)


    def get_with_value(self, namespace, type_name, field_name, value):
        view_namespace = namespace.replace("/", "_")
        view_name = "%s/%s_%s" % (view_namespace, type_name, field_name)
        return self.query_view(view_name, key=value)

    def get_all_type(self, namespace, type_name):
        return self.query_view("raw/all", key=type_name)


    def set_object(self, obj, rev=None):

        return self._set(obj.key, obj._properties, rev=rev)


    @http_backoff
    def _set(self, key, input_value, rev=None, backoff=False):

        value = self.data_adapter.serialise(input_value)

        if self.read_only:
            raise Exception("This db is read only")

        if "type" in value:
            self._check_uniqueness(key, value)

        if self.validator is not None:
            if "namespace" in value and not "schema" in value:
                schema_id = self.validator.schema_id_for(value["namespace"], value["type"])
                if schema_id is not None:
                    value["schema"] = schema_id
            try:
                self.validator.validate(value)
            except jsonschema.ValidationError as e:
                raise FamValidationError(e)


        value["_id"] = key
        if rev:
            value["_rev"] = rev

        url = "%s/%s/%s" % (self.db_url, self.db_name, key)
        rsp = self.session.put(url, data=json.dumps(value,
                                                    indent=4,
                                                    sort_keys=True,
                                                    default=object_default),
                               headers={"Content-Type": "application/json", "Accept": "application/json"})

        if rsp.status_code == 200 or rsp.status_code == 201:
            if rsp.content:
                value["_rev"] = rsp.json()["rev"]

            return ResultWrapper.from_couchdb_json(self.data_adapter.deserialise(value))
        else:
            raise FamResourceConflict("Unknown Error setting CBLite doc: %s %s" % (rsp.status_code, rsp.text))


    def _delete(self, key, rev, classname):

        if self.read_only:
            raise Exception("This db is read only")

        rsp = self.session.delete("%s/%s/%s?rev=%s" % (self.db_url, self.db_name, key, rev))
        if rsp.status_code == 200 or rsp.status_code == 202:
            return
        raise FamResourceConflict("Unknown Error deleting cb doc: %s %s" % (rsp.status_code, rsp.text))


    def _wrapper_from_view_json(self, as_json):
        return ResultWrapper.from_couchdb_view_json(self.data_adapter.deserialise(as_json))


    def _encode_for_view_query(self, kwargs):
        encoded = {}
        for k, v in kwargs.items():
            encoded[k] = json.dumps(v) if k in JSON_KEY_STRINGS else v
        return encoded


    # @ensure_views
    def view(self, name, raw=False, **kwargs):
        design_doc_id, view_name = name.split("/")

        url = self.VIEW_URL % (self.db_url, self.db_name, design_doc_id, view_name)
        rsp = self.session.get(url, params=self._encode_for_view_query(kwargs))

        if rsp.status_code == 200:
            results = rsp.json()
            if raw:
                return results
            rows = results["rows"]
            return [self._wrapper_from_view_json(row) for row in rows]

        raise FamViewError("Unknown Error view cb doc: %s %s %s" % (rsp.status_code, rsp.text, url))


    def authenticate(self):
        pass


    @auth
    def _changes(self, since=None, channels=None, limit=1000, feed=None, timeout=None, filter=None):
        url = "%s/%s/_changes" % (self.db_url, self.db_name)
        params = {"include_docs":"true"}
        if since is not None:
            params["since"] = json.dumps(since)
        if filter is not None and channels is not None:
            raise Exception("you can't specify both filter and channels")
        if filter is not None:
            params["filter"] = filter
        if channels is not None:
            params["filter"] = "sync_gateway/bychannel"
            params["channels"] = ",".join(channels)
        if limit is not None:
            params["limit"] = limit
        if feed is not None:
            params["feed"] = feed
            if feed in ("longpoll", "continuous"):
                if timeout is None:
                    params["timeout"] = 60000
                else:
                    params["timeout"] = timeout
        rsp = self.session.get(url, params=params, cookies=self.cookies)
        # rsp = self.session.get(url, cookies=self.cookies)
        if rsp.status_code == 200:
            results = rsp.json()
            last_seq = results.get("last_seq")
            rows = results.get("results")
            return last_seq, [ResultWrapper.from_couchdb_json(self.data_adapter.deserialise(row["doc"])) for row in rows if "doc" in row.keys() and row["doc"].get(TYPE_STR) is not None]
        if rsp.status_code == 404:
            return None, None
        if rsp.status_code == 403:
            raise FamDbAuthException()
        raise Exception("Unknown Error getting CB doc: %s %s" % (rsp.status_code, rsp.text))


    def clear_all_replications(self):
        return

        url = "{}/_config/replicator/db".format(self.db_url)
        old_replicator_db = self.session.get(url).json()
        new_replicator_db = "replicator_a" if old_replicator_db != "replicator_a" else "replicator_b"
        rsp = self.session.put("{}/{}".format(self.db_url, new_replicator_db))

        if rsp.status_code == 201 or rsp.status_code == 412:
            #set the new db
            self.session.put(url, data=json.dumps(new_replicator_db))
            #delete the old one
            if old_replicator_db != "_replicator":
                rsp = self.session.delete("%s/%s" % (self.db_url, old_replicator_db))
            return new_replicator_db
        else:
           raise Exception("failed to create new replication db")


    def sync_both_continuous(self):

        self.create_sync_up_continuous(self.replicator_db)
        self.create_sync_down_continuous(self.replicator_db)


    def create_sync_down_continuous(self, replicator_db):

        if self.remote_url is None:
            raise Exception("can't sync up nowhere")

        attrs = {"create_target": False,
                 "source": self.remote_url,
                 "target": self.db_name,
                 "continuous": True,
                 "_id": "flotsam_sync_down"
                 }

        headers = {"Content-Type": "application/json"}

        rsp = self.session.post("%s/%s" % (self.db_url, replicator_db), data=json.dumps(attrs), headers=headers)
        if rsp.status_code < 300:
            print("sync down created")
            return

        raise Exception("Error creating sync down: %s %s" % (rsp.status_code, rsp.text))



    def create_sync_up_continuous(self, replicator_db):

        if self.remote_url is None:
            raise Exception("can't sync up nowhere")

        attrs = {"create_target": False,
                 "source": self.db_name,
                 "target": self.remote_url,
                 "continuous": True,
                 "_id": "flotsam_sync_up"
                 }

        headers = {"Content-Type": "application/json"}

        rsp = self.session.post("%s/%s" % (self.db_url, replicator_db), data=json.dumps(attrs), headers=headers)
        if rsp.status_code < 300:
            print("sync_up created")
            return
        raise Exception("Error creating sync up to remote: %s %s" % (rsp.status_code, rsp.text))



    def sync_up(self):

        if self.remote_url is not None:
            attrs = {"create_target": False,
                     "source": self.db_name,
                     "target": self.remote_url}

            headers = {"Content-Type": "application/json"}

            rsp = self.session.post("%s/_replicate" % self.db_url, data=json.dumps(attrs), headers=headers)
            if rsp.status_code < 300:
                return
            raise Exception("Unknown Error syncing up to remote: %s %s" % (rsp.status_code, rsp.text))


    def sync_down(self, continuous=False):

        if self.remote_url is not None:

            attrs = {"create_target": False,
                     "source": self.remote_url,
                     "target": self.db_name}

            headers = {"Content-Type": "application/json"}

            rsp = self.session.post("%s/_replicate" % self.db_url, data=json.dumps(attrs), headers=headers)
            if rsp.status_code < 300:
                return
            raise Exception("Unknown Error syncing up to remote: %s %s" % (rsp.status_code, rsp.text))


    def flush(self):

        rsp = self.session.post("%s/%s/_ensure_full_commit" % (self.db_url, self.db_name))
        if rsp.status_code <= 201:
                return
        raise Exception("Unknown Error _ensure_full_commit in remote: %s %s" % (rsp.status_code, rsp.text))


    def ensure_design_doc(self, key, doc):
        if self.read_only:
            raise Exception("This db is read only")

        # first put it into dev
        dev_key = key.replace("_design/", "_design/dev_")
        dev_doc = copy.deepcopy(doc)
        dev_doc["_id"] = dev_key

        previous_dev = self._get(dev_key)

        # print "self.db_url: ", self.db_url, self.db_name

        self._set(dev_key, dev_doc, rev=None if previous_dev is None else previous_dev.rev)

        # then get it back again to compare
        existing = self._get(key)
        existing_dev = self._get(dev_key)

        if existing == existing_dev:
            print("************  designs up to date ************")
        else:
            print("************  updating designs ************")
            print("new_design: ", doc)
            self._set(key, doc, rev=None if existing is None else existing.rev)



    def _raw_design_doc(self):

        design_doc = {
            "views": {
                "all": {
                    "map": "function(doc) {emit(doc.type, doc);}"
                }
            }
        }

        return design_doc

    def update_designs(self):

        ## simple type index
        doc = self._raw_design_doc()
        key = "_design/raw"

        doc["_id"] = key

        self.ensure_design_doc(key, doc)


        ## relational indexes
        for namespace_name, namespace in self.mapper.namespaces.items():

            view_namespace = namespace_name.replace("/", "_")
            key = "_design/%s" % view_namespace

            doc = self.mapper.get_design(namespace, namespace_name, self.FOREIGN_KEY_MAP_STRING)
            doc["_id"] = key
            self.ensure_design_doc(key, doc)

        ## extra indexes
        for doc in self.mapper.extra_design_docs():
            key = doc["_id"]
            self.ensure_design_doc(key, doc)


    def get_unique_instance(self, namespace, type_name, field_name, value):

        view_namespace = namespace.replace("/", "_")
        view_name = "%s/%s_%s" % (view_namespace, type_name, field_name)
        all_existing = self.query_view(view_name, key=value)
        all_non_null = [o for o in all_existing if getattr(o, field_name, None)]
        how_many_existing = len(all_non_null)

        if how_many_existing > 1:
            raise FamUniqueError("more than {} with a {} of value {}".format(type_name, field_name, value))
        elif how_many_existing == 1:
            return all_non_null[0]
        else:
            return None


    def _check_uniqueness(self, key, value):


        type_name = value["type"]
        namespace = value["namespace"]

        cls = self.mapper.get_class(type_name, namespace)

        for field_name in value:
            if field_name in cls.fields:
                field = cls.fields[field_name]
                if field.unique:
                    this_value = value.get(field_name)
                    if this_value is None:
                        continue

                    type_name = cls._type_with_ref(field_name)
                    existing = self.get_unique_instance(namespace, type_name, field_name, this_value)

                    if existing and existing.key != key:
                        raise FamUniqueError("more than {} with a {} of value {}".format(type_name, field_name, this_value))



from fam.exceptions import *

from fam.database.firestore_adapter import FirestoreDataAdapter


class FirestoreSyncer(object):

    def __init__(self, couchdb_wrapper, firestore_wrapper, batch_size=100, since_in_db=False):

        self.couchdb_wrapper = couchdb_wrapper
        self.firestore_wrapper = firestore_wrapper
        self.queries = []
        self.doc_refs = []
        self.batch_size = batch_size
        self.data_adapter = FirestoreDataAdapter()
        self.types_to_sync_up = None
        self._since = 0
        self.since_rev = None
        self.since_in_db = since_in_db


    def add_query(self, query):
        self.queries.append(query)


    def add_doc_ref(self, doc_ref):
        self.doc_refs.append(doc_ref)

    def add_snapshot(self, snapshot):
        item = self.firestore_wrapper.value_from_snapshot(snapshot)
        update_time = snapshot.update_time
        item["update_seconds"] = update_time.seconds
        item["update_nanos"] = update_time.nanos
        try:
            self.couchdb_wrapper._set(item["_id"], item)
            return item
        except FamResourceConflict as e:
            existing = self.couchdb_wrapper._get(item["_id"])
            if update_time.seconds > existing.value["update_seconds"] or \
                    (update_time.seconds == existing.value["update_seconds"] and
                     update_time.nanos > existing.value["update_nanos"]):
                item["_rev"] = existing.rev
                self.couchdb_wrapper._set(item["_id"], item)
                return item

        return None

    def sync_down(self):

        items = []

        for query in self.queries:
            for snapshot in self.firestore_wrapper.query_snapshots(query, batch_size=self.batch_size):
                item = self.add_snapshot(snapshot)
                if item is not None:
                    items.append(item)

        for doc_ref in self.doc_refs:
            snapshot = doc_ref.get()
            item = self.add_snapshot(snapshot)
            if item is not None:
                items.append(item)

        #self.since = self.couchdb_wrapper.info()["update_seq"]
        return items


    def sync_up(self):
        fs = self.firestore_wrapper.db

        while True:
            last_seq, changes = self.couchdb_wrapper._changes(since=self.since, limit=self.batch_size)

            if changes:
                count = 0
                batch = fs.batch()
                for item in changes:
                    type_name = item.value["type"]
                    if self.types_to_sync_up is None or type_name in self.types_to_sync_up:
                        value = item.value
                        if "update_nanos" in value:
                            del value["update_nanos"]
                        if "update_seconds" in value:
                            del value["update_seconds"]
                        value["_id"] = item.key
                        serialised_value = self.data_adapter.serialise(value)
                        # del serialised_value["_id"]
                        del serialised_value["type"]
                        del serialised_value["namespace"]
                        batch.set(fs.collection(type_name).document(item.key), serialised_value)
                        count += 1

                batch.commit()
                self.since = last_seq
                print("synced up %s items" % count)
            else:
                print("nothing to sync")
                break


    def get_since(self):

        if self.since_in_db:
            since_result = self.couchdb_wrapper._get("sync_since")
            if since_result is not None:
                self.since_rev = since_result.rev
                return since_result.value["since"]
            else:
                return 0
        else:
            return self._since


    def set_since(self, since):

        if self.since_in_db:
            since_result = self.couchdb_wrapper._get("sync_since")
            if since_result is not None:
                self.since_rev = since_result.rev
            value = {"since": since}
            if self.since_rev is not None:
                value["_rev"] = self.since_rev
            self.couchdb_wrapper._set("sync_since", value)
        else:
            self._since = since


    since = property(get_since, set_since)



import unittest
from decimal import Decimal
from fractions import Fraction
from fam.extra_types.lat_long import LatLong
from google.cloud.firestore_v1beta1 import GeoPoint
from fam.tests.test_couchdb.config import *

import os
import datetime
import pytz

from fam.database.couchdb_adapter import CouchDBDataAdapter



import fam
from fam.exceptions import *
from fam.tests.models.test04 import Fish

from fam.database import CouchDBWrapper
from fam.mapper import ClassMapper

SECRETS_DIR = os.path.join(os.path.dirname(fam.__file__), "tests", "secrets")
DATA_DIR = os.path.join(os.path.dirname(fam.__file__), "tests", "data")


class TestSerialisation(unittest.TestCase):

    """
    str
    list
    dict
    bool
    float
    int
    Decimal
    Fraction
    LatLong
    datetime
    bytes
    unicode/str utf8
    """


    def setUp(self):
        self.adapter = CouchDBDataAdapter()


    def test_serialise_basic_types(self):

        doc = { "name": "Paul",
                "age": 53,
                "height": 5.9,
                "is_stoopid": True,
                "children": ["Sol", "Jake"],
                "favorites": {"drink": "coffee", "food": "egg and chips"}
        }

        serialised = self.adapter.serialise(doc)

        # these should all pass through unchanged
        self.assertEquals(doc, serialised)


    def test_serialise_numerics(self):

        doc = {"name": "Paul",
               "age": 53,
               "height": Decimal("5.9"),
               "fraction": Fraction(1, 2),
               "is_stoopid": True,
               "children": ["Sol", "Jake"],
               "favorites": {"drink": "coffee", "food": Decimal("4.2")}
               }

        serialised = self.adapter.serialise(doc)


        self.assertEquals(serialised["height"], "::decimal::5.9")
        self.assertEquals(serialised["fraction"], "::fraction::1/2")
        self.assertEquals(serialised["favorites"]["food"], "::decimal::4.2")


    def test_serialise_latlong_datetime(self):

        birthday = datetime.datetime(1964, 12, 5)

        doc = { "name": "Paul",
                "age": 53,
                "height": 5.9,
                "is_stoopid": True,
                "location": LatLong(51.5102213, -0.1178892),
                "birthday": birthday,
                "children": ["Sol", "Jake"],
                "favorites": {"drink": "coffee", "food": "egg and chips"}
        }

        serialised = self.adapter.serialise(doc)


        self.assertTrue(isinstance(serialised["location"], str))
        self.assertTrue(isinstance(serialised["birthday"], str))
        self.assertEquals(serialised["birthday"], "::datetime::1964-12-05T00:00:00Z")
        self.assertEquals(serialised["location"], "::latlong::51.5102213,-0.1178892")




    def test_serialise_bytes(self):

        doc = { "name": "Paul",
                "age": 53,
                "height": 5.9,
                "is_stoopid": True,
                "birthday": b"aq3restdyrgvdhrjb",
                "children": ["Sol", "Jake"],
                "favorites": {"drink": "coffee", "food": "egg and chips"}
        }

        serialised = self.adapter.serialise(doc)

        self.assertTrue(isinstance(serialised["birthday"], str))


class TestDeSerialisation(unittest.TestCase):




    def setUp(self):
        self.adapter = CouchDBDataAdapter()


    def test_deserialise_datetime(self):

        birthday = datetime.datetime(1964, 12, 5, 12, tzinfo=pytz.UTC)

        doc = { "name": "Paul",
                "age": 53,
                "height": 5.9,
                "new_datetime": birthday,
                "old_datetime": "1964-12-05T12:00:00Z"

        }

        serialised = self.adapter.serialise(doc)

        # these should all pass through unchanged
        self.assertEquals(serialised["new_datetime"], "::datetime::1964-12-05T12:00:00Z")
        self.assertEquals(serialised["old_datetime"], "1964-12-05T12:00:00Z")

        deserialised = self.adapter.deserialise(serialised)

        self.assertEquals(deserialised["new_datetime"], birthday)
        self.assertEquals(deserialised["old_datetime"], birthday)






class TestDatabase(unittest.TestCase):


    def setUp(self):
        mapper = ClassMapper([Fish])
        self.db = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True)
        self.db.update_designs()





    def test_fish(self):

        loc = LatLong(51.2345, -1.4533)
        birthday = datetime.datetime(1964, 12, 5, tzinfo=pytz.utc)

        image_path = os.path.join(DATA_DIR, "goldfish.jpg")
        with open(image_path, "rb") as f:
            image_data = f.read()

        fish = Fish.create(self.db, name="Nemo",
                           location=loc,
                           born=birthday,
                           length=Decimal("45.7"),
                           edible_fraction=Fraction(1, 3),
                           image=image_data
        )

        fish2 = Fish.get(self.db, fish.key)

        self.assertTrue(type(fish2.location) == LatLong)
        self.assertEqual(fish2.location.latitude, 51.2345)

        self.assertTrue(isinstance(fish2.born, datetime.datetime))
        self.assertEqual(fish2.born, birthday)

        self.assertTrue(isinstance(fish2.length, Decimal))
        self.assertEqual(fish2.length, Decimal("45.7"))

        self.assertTrue(isinstance(fish2.edible_fraction, Fraction))
        self.assertEqual(fish2.edible_fraction, Fraction(1, 3))

        self.assertTrue(isinstance(fish2.image, bytes))
        self.assertEqual(fish2.image, image_data)

        ## test for view query
        fishes = list(Fish.all(self.db))
        fish3 = fishes[0]

        self.assertTrue(type(fish3.location) == LatLong)
        self.assertEqual(fish3.location.latitude, 51.2345)

        self.assertTrue(isinstance(fish3.born, datetime.datetime))
        self.assertEqual(fish3.born, birthday)

        self.assertTrue(isinstance(fish3.length, Decimal))
        self.assertEqual(fish3.length, Decimal("45.7"))

        self.assertTrue(isinstance(fish3.edible_fraction, Fraction))
        self.assertEqual(fish3.edible_fraction, Fraction(1, 3))

        self.assertTrue(isinstance(fish3.image, bytes))
        self.assertEqual(fish3.image, image_data)









import json
import os
import unittest
from fam.exceptions import FamValidationError
from fam.tests.models import test01
from fam.tests.models.test01 import Dog, Cat, Person, JackRussell

from fam.schema.writer import createJsonSchema
from fam.schema.validator import ModelValidator

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(THIS_DIR, "data")

class SchemaBaseTests:

    class SchemaTests(unittest.TestCase):

        db = None

        def test_make_a_schema(self):
            expected = {
                "title": "A Fam object model for class glowinthedark.co.uk/test:Cat",
                "required": ['legs', 'namespace', 'owner_id', 'type'],
                "properties": {
                    "_id": {"type": "string"},
                    "_rev": {"type": "string"},
                    "name": {
                        "type": "string"
                    },
                    "colour": {
                        "type": "string"
                    },
                    "namespace": {
                        "pattern": "glowinthedark.co.uk/test",
                        "type": "string"
                    },
                    "owner_id": {
                        "type": "string"
                    },
                    "_deleted": {
                        "type": "boolean"
                    },
                    "tail": {
                        "type": "boolean"
                    },
                    "legs": {
                        "type": "number"
                    },
                    "type": {
                        "pattern": "cat",
                        "type": "string"
                    },
                    "schema": {
                        "type": "string"
                    },
                    "email": {
                        "type": "string",
                        "pattern": "^([-!#$%&'*+/=?^_`{}|~0-9a-zA-Z]+(\\.[-!#$%&'*+/=?^_`{}|~0-9a-zA-Z]+)*|^\"([\\001-\\010\\013\\014\\016-\\037!#-\\[\\]-\\177]|\\\\[\\001-011\\013\\014\\016-\\177])*\")@(?:[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?\\.)+[a-zA-Z]{2,6}\\.?$"
                    }
                },
                "additionalProperties": False,
                "$schema": "http://json-schema.org/draft-04/schema#",
                "type": "object"
            }

            cat_schema = createJsonSchema(Cat)

            print(json.dumps(cat_schema, indent=4))
            self.maxDiff = None
            self.assertEqual(expected, cat_schema)


        def test_make_a_validator(self):


            validator = ModelValidator()

            validator.add_schema(test01.NAMESPACE, "cat", Cat)
            validator.add_schema(test01.NAMESPACE, "person", Person)

            #add validator to db
            self.db.validator = validator

            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(name="whiskers", owner_id=paul.key, legs=4)
            cat.save(self.db)
            self.assertEqual(cat.owner, paul)
            self.assertEqual(cat.owner.name, "paul")

            cat = Cat(name="puss", owner_id=paul.key)

            self.assertRaises(FamValidationError, cat.save, self.db)


        def test_make_a_validator_from_classes(self):

            validator = ModelValidator(classes=[Cat, Person])

            #add validator to db
            self.db.validator = validator

            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(name="whiskers", owner_id=paul.key, legs=4)
            cat.save(self.db)
            self.assertEqual(cat.owner, paul)
            self.assertEqual(cat.owner.name, "paul")

            cat = Cat(name="puss", owner_id=paul.key)

            self.assertRaises(FamValidationError, cat.save, self.db)


        def test_make_a_validator_from_modules(self):

            validator = ModelValidator(modules=[test01])

            #add validator to db
            self.db.validator = validator
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(name="whiskers", owner_id=paul.key, legs=4)
            cat.save(self.db)
            self.assertEqual(cat.owner, paul)
            self.assertEqual(cat.owner.name, "paul")
            cat = Cat(name="puss", owner_id=paul.key)

            self.assertRaises(FamValidationError, cat.save, self.db)



import sys
import os
import time
import subprocess
import unittest
import time
from fam.exceptions import *
from fam.database import SyncGatewayWrapper
from fam.mapper import ClassMapper
from fam.tests.test_sync_gateway.config import *
from fam.tests.models.test01 import Dog, Cat, Person, JackRussell, Monkey, Monarch, Monster
from fam.tests.models.test01 import GenericObject, Dog, Cat, Person, JackRussell, Monkey, Monarch, NAMESPACE
from fam.tests.common import common_test_classes

current_module = sys.modules[__name__]

TEST_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DATA_PATH = os.path.join(TEST_DIR, "common", "data")


class BasicTests(unittest.TestCase):

    def setUp(self):
        cmd = "{} -log=* -url walrus: ".format(SYNC_GATEWAY_PATH)
        # print cmd

        time.sleep(0.25)
        self.gateway = subprocess.Popen(cmd, shell=True)
        time.sleep(0.25)
        filepath = os.path.join(DATA_PATH, "animal_views.js")
        mapper = ClassMapper([Dog, Cat, Person, JackRussell, Monkey, Monarch, Monster], designs=[filepath])

        url = "http://%s:%s" % (SYNC_GATEWAY_ADMIN_HOST, SYNC_GATEWAY_ADMIN_PORT)
        self.db = SyncGatewayWrapper(mapper, url, SYNC_GATEWAY_NAME)
        self.db.update_designs()
        super(self.__class__, self).setUp()

    def tearDown(self):
        # stop the gateway
        self.gateway.kill()

    # Test the app
    def test_app(self):
        self.assertNotEqual(self.db, None)


    def test_make_an_object(self):
        dog = Dog(name="woofer")
        self.assertEqual(dog.namespace, NAMESPACE)
        self.assertEqual(dog.type, "dog")
        self.assertEqual(dog.name, "woofer")
        self.assertEqual(dog.__class__, Dog)


    def test_make_an_object2(self):
        dog = Dog(name="woofer")
        self.assertEqual(dog.talk(), "woof")


    def test_make_a_sub_object(self):
        jack = JackRussell()
        self.assertEqual(jack.talk(), "Yap")
        jack.name = "jack"
        jack.age = 12
        self.db.put(jack)


    def test_make_an_object_saved(self):
        dog = Dog(name="woofer")
        dog.save(self.db)
        self.assertEqual(dog.namespace, NAMESPACE)
        self.assertEqual(dog.type, "dog")
        self.assertEqual(dog.name, "woofer")
        self.assertEqual(dog.__class__, Dog)
        self.assertNotEqual(dog.key, None)
        # self.assertTrue(dog.schema.startswith("http://glowinthedark.co.uk/test/dog"))


    def test_make_an_object_double_dispatch_saved(self):
        dog = Dog(name="woofer")
        self.db.put(dog)
        self.assertEqual(dog.namespace, NAMESPACE)
        self.assertEqual(dog.type, "dog")
        self.assertEqual(dog.name, "woofer")
        self.assertEqual(dog.__class__, Dog)
        self.assertNotEqual(dog.key, None)


    def test_make_an_object_with_additional_properties(self):
        dog = Dog(name="woofer", collar="leather")
        dog.home = "hackney"
        self.db.put(dog)
        self.assertEqual(dog.home, "hackney")

    def test_fail_with_additional_properties(self):

        def wrong_monkey():
            return Monkey(name="bonzo", collar="leather")

        def collar_monkey(monkey):
            monkey.collar = "leather"

        self.assertRaises(FamValidationError, wrong_monkey)
        monkey = Monkey(name="bonzo")
        self.assertRaises(FamValidationError, collar_monkey, monkey)


    def test_make_an_object_saved_cas(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        cat.save(self.db)
        self.assertEqual(cat.namespace, NAMESPACE)
        self.assertEqual(cat.type, "cat")
        self.assertNotEqual(cat.rev, None)


    def test_get_cat(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        cat.save(self.db)
        another = Cat.get(self.db, cat.key)
        self.assertEqual(cat, another)


    def test_ref_to(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        cat.save(self.db)
        self.assertEqual(cat.owner, paul)
        self.assertEqual(cat.owner.name, "paul")


    def test_ref_to_with_object(self):
        paul = Person(name="paul")
        self.db.put(paul)
        cat = Cat(name="whiskers", owner=paul, legs=2)
        cat.save(self.db)
        self.assertEqual(cat.owner, paul)
        self.assertEqual(cat.owner.name, "paul")
        catback = self.db.get(cat.key)
        self.assertEqual(cat, catback)


    def test_ref_from(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        cat.save(self.db)
        cat2 = Cat(name="puss", owner_id=paul.key, legs=2)
        cat2.save(self.db)
        cats = list(paul.cats)
        self.assertEqual(len(cats), 2)
        cats = list(paul.cats)
        self.assertTrue(cats[0] == cat or cats[1] == cat)


    def test_ref_from_multiple_index(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        cat.save(self.db)
        dog = Dog(name="fly", owner=paul)
        self.db.put(dog)
        cats = list(paul.cats)
        self.assertEqual(len(cats), 1)
        dogs = list(paul.dogs)
        self.assertEqual(len(dogs), 1)
        animals = list(paul.animals)
        self.assertEqual(len(animals), 2)


    def test_refs_with_inheritance(self):
        paul = Person(name="paul")
        paul.save(self.db)
        jack = JackRussell()
        jack.owner_id = paul.key
        jack.name = "jack"
        jack.save(self.db)

        self.assertEqual(list(paul.dogs)[0], jack)


    def test_refs_with_other_inheritance(self):
        paul = Monarch(name="paul")
        self.db.put(paul)
        jack = Dog()
        jack.owner = paul
        jack.name = "jack"
        self.db.put(jack)
        dogs = list(paul.dogs)
        self.assertEqual(dogs[0], jack)


    def test_delete_cat_dd(self):
        paul = Person(name="paul")
        self.db.put(paul)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        self.db.put(cat)
        key = cat.key
        another = self.db.get(cat.key)
        self.assertEqual(cat, another)
        self.db.delete(cat)
        revivedcat = self.db.get(key)
        self.assertEqual(revivedcat, None)


    def test_delete_dog_to_refs(self):
        paul = Person(name="paul")
        paul.save(self.db)
        dog = Dog(name="rufus", owner_id=paul.key)
        dog.save(self.db)
        key = dog.key
        dog2 = Dog(name="fly", owner_id=paul.key)
        dog2.save(self.db)
        self.assertTrue(dog2.owner is not None)
        key = paul.key
        dog.delete(self.db)
        revivedpaul = self.db.get(key)
        self.assertTrue(revivedpaul is None)
        refresheddog2 = Dog.get(self.db, dog2.key)
        self.assertTrue(refresheddog2.owner is None)


    def test_delete_cat_refs(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        cat.save(self.db)
        key = cat.key
        cat2 = Cat(name="puss", owner_id=paul.key, legs=2)
        cat2.save(self.db)

        cats = list(paul.cats)

        # revivedcat1 = self.db.get(key)
        #
        # self.assertTrue(revivedcat1 is not None)
        #
        # paul.delete(self.db)
        # revivedcat2 = self.db.get(key)
        # self.assertTrue(revivedcat2 is None)


    def test_delete_dog_refs(self):
        paul = Person(name="paul")
        paul.save(self.db)
        dog = Dog(name="rufus", owner_id=paul.key)
        dog.save(self.db)
        key = dog.key
        dog2 = Dog(name="fly", owner_id=paul.key)
        dog2.save(self.db)
        reviveddog1 = self.db.get(key)
        self.assertTrue(reviveddog1 is not None)
        paul.delete(self.db)
        reviveddog2 = self.db.get(key)
        self.assertTrue(reviveddog2 is not None)


    def test_update_cat(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        cat.save(self.db)
        cat.name = "blackie"
        cat.save(self.db)
        self.assertEqual(cat.name, "blackie")
        self.assertEqual(cat._properties["name"], "blackie")
        self.assertFalse("name" in cat.__dict__.keys())


    def test_update_cat_fails(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        cat.save(self.db)
        cat.colour = "black"
        cat.save(self.db)
        def change_colour():
            cat.colour = "white"
        self.assertRaises(FamImmutableError, change_colour)


    def setcatfood(self):
        self.cat.food = "biscuits"


    def test_update_catfood(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        cat.save(self.db)
        self.assertRaises(Exception, self.setcatfood)


    def test_update_dogfood(self):
        dog = Dog(name="fly")
        dog.food = "biscuits"
        dog.save(self.db)


    def test_update_fails_without_rev(self):
        dog = Dog(name="fly")
        dog.save(self.db)
        dog.rev = None
        dog.name = "jess"
        self.assertRaises(FamResourceConflict, self.db.put, dog)


    def test_update_works_without_rev(self):
        monkey = Monkey(name="fly")
        monkey.save(self.db)
        monkey.rev = None
        monkey.name = "jess"
        self.db.put(monkey)

    def test_uniqueness(self):

        paul = Person(name="paul")
        self.db.put(paul)
        dog1 = Dog(name="rufus", owner_id=paul.key, kennel_club_membership="123456")

        dog1.save(self.db)
        # time.sleep(1)


        # print dog1.as_json()
        # dog2 = Dog(name="fly", owner_id=paul.key, kennel_club_membership="123456")
        # print dog2.as_json()
        # # self.db.put(dog2)
        # self.assertRaises(FamUniqueError, self.db.put, dog2)
        # # print "*********** end ***************"

class RefNameTests(unittest.TestCase):

    # this test that refs whos name doesnt end with _id fail
    def test_misnamed_ref_to_fails(self):

        def duff_import():
            from fam.tests.models import test02
            print(test02)

        self.assertRaises(FamError, duff_import)


class FamDbConnectionException(Exception):
    pass

class FamResourceConflict(Exception):
    pass

class FamViewError(Exception):
    pass

class FamValidationError(Exception):
    pass

class FamImmutableError(Exception):
    pass

class FamUniqueError(Exception):
    pass

class FamWriteError(Exception):
    pass

class FamError(Exception):
    pass



import unittest

from fam.database import CouchbaseWrapper
from fam.tests.models.test01 import Dog, Cat, Person, JackRussell, NAMESPACE
from fam.mapper import ClassMapper

COUCHBASE_HOST = "127.0.0.1"
COUCHBASE_BUCKET = "test"

class CouchbaseModelTests(object):


    def setUp(self):
        mapper = ClassMapper([Dog, Cat, Person, JackRussell])
        self.db = CouchbaseWrapper(mapper, COUCHBASE_HOST, COUCHBASE_BUCKET, read_only=False)
        self.db.update_designs()

    def tearDown(self):
        pass

   # Test the app
    def test_app(self):
        self.assertNotEqual(self.db, None)


    def test_make_an_object(self):
        dog = Dog(name="woofer")
        self.assertEqual(dog.namespace, NAMESPACE)
        self.assertEqual(dog.type, "dog")
        self.assertEqual(dog.name, "woofer")
        self.assertEqual(dog.__class__, Dog)


    def test_make_an_object2(self):
        dog = Dog(name="woofer")
        self.assertEqual(dog.talk(), "woof")


    def test_make_an_sub_object(self):
        jack = JackRussell()
        self.assertEqual(jack.talk(), "Yap")
        jack.name = "jack"
        jack.age = 12


    def test_make_an_object_saved(self):
        dog = Dog(name="woofer")
        dog.save(self.db)
        self.assertEqual(dog.namespace, NAMESPACE)
        self.assertEqual(dog.type, "dog")
        self.assertEqual(dog.name, "woofer")
        self.assertEqual(dog.__class__, Dog)
        self.assertNotEqual(dog.key, None)


    def test_make_an_object_saved_cas(self):
        cat = Cat(name="whiskers")
        cat.save(self.db)
        self.assertEqual(cat.namespace, NAMESPACE)
        self.assertEqual(cat.type, "cat")
        self.assertNotEqual(cat.cas, None)


    def test_get_cat(self):
        cat = Cat(name="whiskers")
        cat.save(self.db)
        another = Cat.get(self.db, cat.key)
        self.assertEqual(cat, another)


    def test_ref_to(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key)
        cat.save(self.db)
        self.assertEqual(cat.owner, paul)
        self.assertEqual(cat.owner.name, "paul")



    def test_ref_from(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key)
        cat.save(self.db)
        cat2 = Cat(name="puss", owner_id=paul.key)
        cat2.save(self.db)
        self.assertEqual(len(paul.cats), 2)
        self.assertTrue(paul.cats[0] == cat or paul.cats[1] == cat)


    def test_delete_cat(self):
        cat = Cat(name="whiskers")
        cat.save(self.db)
        key = cat.key
        another = Cat.get(self.db, cat.key)
        self.assertEqual(cat, another)
        cat.delete(self.db)
        revivedcat = self.db.get(key)
        self.assertEqual(revivedcat, None)


    def test_delete_dog_to_refs(self):
        paul = Person(name="paul")
        paul.save(self.db)
        dog = Dog(name="rufus", owner_id=paul.key)
        dog.save(self.db)
        key = dog.key
        dog2 = Dog(name="fly", owner_id=paul.key)
        dog2.save(self.db)
        self.assertNotEqual(dog2.owner, None)
        key = paul.key
        dog.delete(self.db)
        revivedpaul = self.db.get(key)
        self.assertEqual(revivedpaul, None)
        refresheddog2 = Dog.get(self.db, dog2.key)
        self.assertEqual(refresheddog2.owner, None)


    def test_delete_cat_refs(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key)
        cat.save(self.db)
        key = cat.key
        cat2 = Cat(name="puss", owner_id=paul.key)
        cat2.save(self.db)
        revivedcat1 = self.db.get(key)
        self.assertNotEqual(revivedcat1, None)
        paul.delete(self.db)
        revivedcat2 = self.db.get(key)
        self.assertEqual(revivedcat2, None)


    def test_delete_dog_refs(self):
        paul = Person(name="paul")
        paul.save(self.db)
        dog = Dog(name="rufus", owner_id=paul.key)
        dog.save(self.db)
        key = dog.key
        dog2 = Dog(name="fly", owner_id=paul.key)
        dog2.save(self.db)
        reviveddog1 = self.db.get(key)
        self.assertNotEqual(reviveddog1, None)
        paul.delete(self.db)
        reviveddog2 = self.db.get(key)
        self.assertNotEqual(reviveddog2, None)


    def test_update_cat(self):
        cat = Cat(name="whiskers")
        cat.save(self.db)
        cat.name = "blackie"
        cat.save(self.db)
        self.assertEqual(cat.name, "blackie")
        self.assertEqual(cat._properties["name"], "blackie")
        self.assertFalse("name" in cat.__dict__.keys())


    def setcatfood(self):
        self.cat.food = "biscuits"


    def test_update_catfood(self):
        self.cat = Cat(name="whiskers")
        self.cat.save(self.db)
        self.assertRaises(Exception, self.setcatfood)


    def test_update_dogfood(self):
        dog = Dog(name="fly")
        dog.food = "biscuits"
        dog.save(self.db)


    def test_all(self):
        dog = Dog(name="fly")
        dog.save(self.db)
        all = Dog.all(self.db)
        self.assertEqual(len(all), 1)




from fam.database.base import BaseDatabase
from fam.buffer.write_buffer import FamWriteBuffer
from .null import NullDatabase


class MockDatabase(BaseDatabase):
    database_type = "mock"
    check_on_save = False

    def __init__(self, mapper):
        self.mapper = mapper
        null_db = NullDatabase(mapper)
        self.buffer = FamWriteBuffer(null_db)


    def set_object(self, obj, rev=None):
        return self.buffer.put(obj)

    def get(self, key, class_name=None):
        return self.buffer.get(key)

    def _delete(self, key, rev, classname):
        return self.buffer.delete_key(key)

    def delete_key(self, key):
        return self.buffer.delete_key(key)

    def query_view(self, view_name, **kwargs):
        return self.buffer.query_view(view_name, **kwargs)

    def get_refs_from(self, namespace, type_name, name, key, field):
        return self.buffer.get_refs_from(namespace, type_name, name, key, field)



from fam.blud import (GenericObject,
                      StringField,
                      ReferenceFrom,
                      ReferenceTo,
                      BoolField,
                      NumberField,
                      DictField,
                      ObjectField,
                      ListField,
                      LatLongField,
                      DateTimeField,
                      FractionField,
                      DecimalField,
                      BytesField)


NAMESPACE = "glowinthedark.co.uk/test"



class House(GenericObject):

    fields = {
        "name": StringField(),
        "location": LatLongField(),
        }


class Fence(GenericObject):

    fields = {
        "name": StringField(),
        "boundary": ListField()
        }


class Fish(GenericObject):

    fields = {
        "name": StringField(),
        "location": LatLongField(),
        "born": DateTimeField(),
        "length": DecimalField(),
        "edible_fraction": FractionField(),
        "image": BytesField()
        }



RESERVED_PROPERTY_NAMES = ("key", "rev", "_properties", "_db")

# NAMESPACE_STR = "namespace"
NAMESPACE_STR = "namespace"
TYPE_STR = "type"

FIELD_REQUIRED = "required"
FIELD_UPDATE_ACL = "update_acl"



import os
import shutil
import unittest
from fam.database import CouchDBWrapper
from fam.mapper import ClassMapper
from fam.schema.validator import ModelValidator
from fam.exceptions import FamValidationError
from fam.tests.test_couchdb.config import *
from fam.tests.models import test01
from fam.tests.models.test01 import Dog, Cat, Person, JackRussell
from fam.blud import StringField

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(THIS_DIR, "data")
DESIGN_PATH = os.path.join(DATA_PATH, "design_ref.json")

class MapperValidationTests(unittest.TestCase):

    def test_make_a_validator(self):

        mapper = ClassMapper([Dog, Cat, Person, JackRussell])
        validator = ModelValidator(classes=[Dog, Cat, Person, JackRussell])
        self.db = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True, validator=validator)
        self.db.update_designs()

        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=4)
        cat.save(self.db)
        self.assertEqual(cat.owner, paul)
        self.assertEqual(cat.owner.name, "paul")

        cat = Cat(name="puss", owner_id=paul.key)

        self.assertRaises(FamValidationError, cat.save, self.db)


    def test_make_a_validator_from_modules(self):

        mapper = ClassMapper([], modules=[test01])
        validator = ModelValidator(classes=[Dog, Cat, Person, JackRussell])
        self.db = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True, validator=validator)
        self.db.update_designs()

        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=4)
        cat.save(self.db)
        self.assertEqual(cat.owner, paul)
        self.assertEqual(cat.owner.name, "paul")

        #missing legs
        cat = Cat(name="puss", owner_id=paul.key)

        self.assertRaises(FamValidationError, cat.save, self.db)

        #additional properties
        def failing_cat():
            cat = Cat(name="puss", owner_id=paul.key, legs=2, collar="green")

        self.assertRaises(FamValidationError, failing_cat)
        dog = Dog(name="fly")
        self.db.put(dog)
        dog.tail = "long"
        self.db.put(dog)

        # print dog.as_json()


    def test_included_refs_from_in_validator(self):
        mapper = ClassMapper([], modules=[test01])
        validator = ModelValidator(modules=[test01])
        self.db = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True, validator=validator)
        self.db.update_designs()

        paul = Person(name="paul")
        paul.save(self.db)

        paul_id = paul.key
        cat = Cat(name="whiskers", owner_id=paul.key, legs=4)
        cat.save(self.db)
        self.assertEqual(cat.owner, paul)
        self.assertEqual(cat.owner.name, "paul")

        paul = Person.get(self.db, paul_id)

        paul.save(self.db)



    def test_string_format(self):

        mapper = ClassMapper([], modules=[test01])
        validator = ModelValidator(modules=[test01])
        self.db = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True, validator=validator)
        self.db.update_designs()

        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=4)
        cat.save(self.db)
        self.assertEqual(cat.owner, paul)
        self.assertEqual(cat.owner.name, "paul")


        cat = Cat(name="puss", owner_id=paul.key, legs=3, email="paul@glowinthedark.co.uk")
        cat.save(self.db)
        cat.email = "paulglowinthedark.co.uk"
        self.assertRaises(FamValidationError, self.db.put, cat)




import sys
import datetime
from fam.extra_types.lat_long import LatLong
from fractions import Fraction
from decimal import Decimal
from google.cloud.firestore_v1beta1 import GeoPoint

from .base_adapter import BaseDataAdapter


if sys.version_info[0] < 3:
    PYTHON_VERSION = 2
else:
    PYTHON_VERSION = 3


class FirestoreDataAdapter(BaseDataAdapter):


    def serialise_lat_long(self, lat_long):
        return GeoPoint(latitude=lat_long.latitude, longitude=lat_long.longitude)


    def serialise_date_time(self, date_time):
        return date_time


    def serialise_bytes(self, btes):
        return btes


    def _deserialise_walk(self, node):

        if isinstance(node, dict):
            for k, v in node.items():
                node[k] = self._deserialise_walk(v)
            return node
        if isinstance(node, list):
            return [self._deserialise_walk(v) for v in node]
        if isinstance(node, GeoPoint):
            return LatLong(latitude=node.latitude, longitude=node.longitude)
        if self.is_a_string(node):
            if node.startswith("::fraction::"):
                stripped = node[len("::fraction::"):]
                num, denom = stripped.split("/")
                return Fraction(int(num), int(denom))
            if node.startswith("::decimal::"):
                stripped = node[len("::decimal::"):]
                return Decimal(stripped)
            return node
        if isinstance(node, datetime.datetime):
            return node
        if isinstance(node, bytes):
            return node
        if isinstance(node, bool):
            return node
        if self.is_a_number(node):
            return node
        if node is None:
            return None

        raise Exception("FirestoreDataAdapter can't deserialise this value: %s" % node)



import sys
import datetime
from copy import deepcopy
from decimal import Decimal
from fractions import Fraction

if sys.version_info[0] < 3:
    PYTHON_VERSION = 2
else:
    PYTHON_VERSION = 3


__all__ = [
    "BoolField",
    "NumberField",
    "StringField",
    "ListField",
    "DictField",
    "ObjectField",
    "LatLongField",
    "DateTimeField",
    "BytesField",
    "DecimalField",
    "FractionField",
    "ReferenceTo",
    "ReferenceFrom",
    "EmailField",
]

from .constants import *
from .exceptions import *
from .extra_types.lat_long import LatLong


class Field(object):
    object = "base"

    def __init__(self, required=False, immutable=False, default=None, unique=False):
        self.required = required
        self.immutable = immutable
        self.default = default
        self.unique = unique
        self._types = []

        if self.default is not None and self.required is True:
            raise FamError("It doesnt really make sense to use both required and default together. Just use default")

    def is_correct_type(self, value):
        value is None or any(isinstance(value, cls) for cls in self._types)

    def get_default(self):
        return self.default

    def __str__(self):
        attr = []
        if self.required:
            attr.append(FIELD_REQUIRED)
        return " ".join(attr)

    as_string = property(__str__)


class BoolField(Field):

    _types = [bool]


class NumberField(Field):

    if PYTHON_VERSION == 3:
        _types = [int, float]
    else:
        _types = [int, float, long]


class StringField(Field):

    if PYTHON_VERSION == 3:
        _types = [str]
    else:
        _types = [basestring]


class ListField(Field):

    _types = [list]

    def __init__(self, item_cls=None, required=False, immutable=False, default=None):
        self.item_cls = item_cls
        super(ListField, self).__init__(required=required, immutable=immutable, default=default)

    def get_default(self):
        return deepcopy(self.default)



class DictField(Field):

    _types = [dict]

    def get_default(self):
        return deepcopy(self.default)



class ObjectField(Field):

    def get_default(self):
        return self.cls()

    def __init__(self, cls, default=None, required=False):
        self.cls = cls
        self._types = [cls]
        super(ObjectField, self).__init__(default=default, required=required)


    # "LatLongField",
    # "DateTimeField",
    # "ByteField",
    # "DecimalField",
    # "FractionField",

class LatLongField(Field):

    _types = [LatLong]

    def get_default(self):
        return LatLong(self.default.latitude, self.default.longitude)


class DateTimeField(Field):
    _types = [datetime.datetime]

    def get_default(self):
        return self.default


class BytesField(Field):

    if PYTHON_VERSION == 3:
        _types = [bytes, bytearray]
    else:
        _types = [bytes, bytearray, str]

    def get_default(self):
        return self.default.copy()


class DecimalField(Field):

    _types = [Decimal]

    def get_default(self):
        return self.default.copy()


class FractionField(Field):
    _types = [Fraction]

    def get_default(self):
        return self.default.copy()


class EmailField(StringField):

    pattern = r"""^([-!#$%&'*+/=?^_`{}|~0-9a-zA-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9a-zA-Z]+)*|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-011\013\014\016-\177])*")@(?:[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?\.)+[a-zA-Z]{2,6}\.?$"""



class ReferenceTo(Field):

    if PYTHON_VERSION == 3:
        _types = [str]
    else:
        _types = [basestring]

    def __init__(self, refns, refcls, required=False, immutable=False, default=None, unique=False,
                 cascade_delete=False):
        self.refns = refns
        self.refcls = refcls
        self.cascade_delete = cascade_delete
        super(ReferenceTo, self).__init__(required, immutable, default, unique)

    def __str__(self):
        attr = []

        attr.append("ns:%s" % self.refns)
        attr.append("resource:%s" % self.refcls)

        if self.required:
            attr.append(FIELD_REQUIRED)
        return " ".join(attr)

    as_string = property(__str__)


class ReferenceFrom(Field):

    def __init__(self, refns, refcls, fkey, required=False, immutable=False, default=None, cascade_delete=False):
        self.refns = refns
        self.refcls = refcls
        self.fkey = fkey
        self.cascade_delete = cascade_delete
        super(ReferenceFrom, self).__init__(required, immutable, default)

    def __str__(self):
        attr = []
        attr.append("ns:%s" % self.refns)
        attr.append("resource:%s" % self.refcls)
        attr.append("key:%s" % self.fkey)
        if self.required:
            attr.append(FIELD_REQUIRED)
        return " ".join(attr)

    as_string = property(__str__)


import unittest
import os
import time
import json
import subprocess
from fam.database import SyncGatewayWrapper
from fam.mapper import ClassMapper
from fam.tests.test_sync_gateway.config import *
from fam.tests.models.acl import Car, Bike, Boat
from fam.acl.writer import write_sync_function, _requirements_from_mapper
from fam.utils import couchbase_utils

from fam.exceptions import *

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(TEST_DIR, "data")

"""
Actors are people who do things

users: a set of named users
roles: a set of people who have any of given roles
owner: an owner as named in the doc
anyone: anyone as long as we know who they are

Actions are things they can do:

create: Create a new document with a new id
update: Make a change to an existing document
delete: Delete an existing document

"""


class testPermissions(unittest.TestCase):

        def setUp(self):

            self.gateway = None
            self.db = None

            self.mapper = ClassMapper([Car, Boat])



        def start_gateway(self, conf_file_path):

            cmd = "{} -log=* {}".format(SYNC_GATEWAY_PATH, conf_file_path)
            print cmd

            time.sleep(0.25)
            self.gateway = subprocess.Popen(cmd, shell=True)
            time.sleep(0.25)

            admin_url = "http://%s:%s" % (SYNC_GATEWAY_ADMIN_HOST, SYNC_GATEWAY_ADMIN_PORT)
            self.admin_db = SyncGatewayWrapper(self.mapper, admin_url, SYNC_GATEWAY_NAME)
            self.admin_db.update_designs()
            self.add_users()


        def add_users(self):

            admin_url = "http://%s:%s" % (SYNC_GATEWAY_ADMIN_HOST, SYNC_GATEWAY_ADMIN_PORT)

            couchbase_utils.add_person_to_gateway(admin_url,
                                                  SYNC_GATEWAY_NAME,
                                                  "paul_id",
                                                  "paul",
                                                  "password1",
                                                  admin_channels=["cars", "paul"])

            couchbase_utils.add_person_to_gateway(admin_url,
                                                  SYNC_GATEWAY_NAME,
                                                  "sol_id",
                                                  "sol",
                                                  "password2",
                                                  admin_channels=["sol"])

            paul_url = "http://paul:password1@%s:%s" % (SYNC_GATEWAY_HOST, SYNC_GATEWAY_PORT)
            self.paul_db = SyncGatewayWrapper(self.mapper, paul_url, SYNC_GATEWAY_NAME)

            sol_url = "http://sol:password2@%s:%s" % (SYNC_GATEWAY_HOST, SYNC_GATEWAY_PORT)
            self.sol_db = SyncGatewayWrapper(self.mapper, sol_url, SYNC_GATEWAY_NAME)


        def tearDown(self):
            # stop the gateway
            if self.gateway is not None:
                self.gateway.kill()


        def test_index_permissions(self):

            expected = {
                "create": {
                    "car": {
                        "owner": True
                    },
                    "boat": {
                        "owner": True,
                        "withoutAccess": True
                    }
                },
                "update": {
                    "car": [
                        {
                            "fields": [
                                "access"
                            ],
                            "role": []
                        },
                        {
                            "owner": True,
                            "fields": [
                                "colour"
                            ]
                        }
                    ],
                    "boat": [
                        {
                            "role": []
                        }
                    ]
                },
                "delete": {
                    "car": {
                        "owner": True
                    },
                    "boat": {
                        "owner": True
                    }
                }
            }



            requirements = _requirements_from_mapper(self.mapper)

            print json.dumps(requirements, indent=4)

            self.assertEqual(expected, requirements)


        def test_write_permissions(self):

            src_path = os.path.join(DATA_PATH, "sync_conf_template")
            dst_path = os.path.join(DATA_PATH, "sync_conf")

            if os.path.exists(dst_path):
                os.remove(dst_path)

            write_sync_function(src_path, dst_path, self.mapper)

            self.start_gateway(dst_path)


        def test_owner_create(self):

            self.test_write_permissions()

            car = Car(colour="red", stars=3, owner_name="paul", channels=["cars", "sol"])
            self.paul_db.put(car)

            car.colour = "green"
            self.paul_db.put(car)
            car.stars = 2
            self.paul_db.put(car)
            self.paul_db.delete(car)

            car2 = Car(colour="green", owner_name="sol", channels=["cars", "sol"])
            self.assertRaises(FamResourceConflict, self.paul_db.put, car2)


        def test_non_owner_permissions(self):

            self.test_write_permissions()

            car = Car(colour="red", stars=3, owner_name="paul", channels=["cars", "sol"])
            self.paul_db.put(car)
            sols_car = self.sol_db.get(car.key)

            ## changing green fails
            sols_car.colour = "green"
            self.assertRaises(FamResourceConflict, self.sol_db.put, sols_car)

            ## changing stars works
            sols_car.colour = "red"
            sols_car.stars = 2
            self.sol_db.put(sols_car)

            # check nn owner cant delete
            self.assertRaises(FamResourceConflict, self.sol_db.delete, sols_car)


        def test_wrong_owner_create_fails(self):

            self.test_write_permissions()
            car = Car(colour="red", stars=3, owner_name="sol", channels=["cars", "sol"])
            self.assertRaises(FamResourceConflict, self.paul_db.put, car)


        def test_create_access(self):
            self.test_write_permissions()
            car1 = Car(colour="red", stars=3, owner_name="sol", channels=["sol"])
            self.sol_db.put(car1)
            self.sol_db.get(car1.key)

            car2 = Car(colour="green", stars=2, owner_name="sol", channels=["paul"])
            self.assertRaises(Exception, self.sol_db.put, car2)
            self.sol_db.get(car2.key)


        def test_change_access_without_permission_fails(self):
            self.test_write_permissions()
            car = Car(colour="red", stars=3, owner_name="paul", channels=["cars"])
            self.paul_db.put(car)
            car.access = ["sol"]
            self.assertRaises(FamResourceConflict, self.paul_db.put, car)


        def test_change_access(self):
            self.test_write_permissions()
            car = Car(key="cars", colour="red", stars=3, owner_name="paul", channels=["cars"])
            self.paul_db.put(car)

            ## sol cant get car
            self.assertRaises(Exception, self.sol_db.get, "cars")

            car.access = ["sol"]
            self.admin_db.put(car)
            self.sol_db.get(car.key)


        def test_no_access(self):
            self.test_write_permissions()
            bike = Bike(wheels = 2)
            self.assertRaises(Exception, self.paul_db.put, bike)


        def test_no_access_admin(self):
            self.test_write_permissions()
            bike = Bike(wheels=2)
            self.admin_db.put(bike)


        def test_own_access_create(self):

            self.test_write_permissions()
            boat_id = "boaty"
            boat = Boat(key=boat_id, name="steve", is_sail=True, owner_name="paul", access=["paul"], channels=[boat_id])
            self.paul_db.put(boat)








## little dance to use patch version if necessary

def is_gevent_monkey_patched():
    try:
        from gevent import monkey
    except ImportError:
        return False
    else:
        return monkey.is_module_patched('__builtin__')

if is_gevent_monkey_patched():
    from gcouchbase import Bucket
else:
    from couchbase.bucket import Bucket

from couchbase.bucket import View
from couchbase.n1ql import N1QLQuery
from couchbase.exceptions import NotFoundError, KeyExistsError


from fam.blud import FamObject
from fam.database.base import BaseDatabase
from fam.database.couchdb import ResultWrapper
from fam.exceptions import *


class CouchbaseWrapper(BaseDatabase):

    def __init__(self, mapper, host, bucket_name, read_only=True):
        connection_str = "couchbase://%s/%s" % (host, bucket_name)
        self.bucket = Bucket(connection_str)
        self.read_only = read_only
        self.mapper = mapper



    def view(self, name, *args, **kwargs):

        design_doc_id, view_name = name.split("/")

        design_name = "_design/%s" % design_doc_id

        # print design_doc_id, view_name

        view = View(self.bucket,
                    design_doc_id,
                           view_name,
                           *args,
                           **kwargs
                           )

        rows_list = list(view)
        # print rows_list
        #
        #
        #
        # keys = rows_list[0].keys()
        # keys.remove("id")
        # keys.remove("cas")
        return [ResultWrapper(row.docid, row.doc.cas, row.value) for row in rows_list]

    def n1ql(self, query, with_revs=False, *args, **kwargs):
        return FamObject.n1ql(self, query, with_revs=with_revs, *args, **kwargs)

    def _get(self, key):
        try:
            result = self.bucket.get(key)
        except NotFoundError as e:
            return None
        return ResultWrapper(key, result.cas, result.value)

    def _set(self, key, value, rev=None):
        if self.read_only:
            raise FamWriteError("You can't write to this database")
        try:
            print("rev", rev)
            if rev is not None:
                result = self.bucket.upsert(key, value, cas=rev)
            else:
                result = self.bucket.upsert(key, value)
            print("result", result)
            return ResultWrapper(key, result.cas, value)
        except KeyExistsError as e:
            raise FamResourceConflict("key alreday exists in couchbase: %s - %s" % (key, e))


    def _n1ql_with_rev(self, query, *args, **kwargs):
        query = N1QLQuery(query, *args, **kwargs)
        rows = self.bucket.n1ql_query(query)
        results = []
        bucket_name = None

        for row in rows:
            if bucket_name is None:
                keys = row.keys()
                keys.remove("id")
                keys.remove("cas")
                bucket_name = keys[0]

            rev = row["_sync"]["rev"]
            results.append(ResultWrapper(row["id"], rev, row[bucket_name]))

        return results


    def _n1ql(self, query, *args, **kwargs):
        query = N1QLQuery(query, *args, **kwargs)
        rows = self.bucket.n1ql_query(query)
        rows_list = list(rows)
        # print "rows_list: ", rows_list
        keys = rows_list[0].keys()
        keys.remove("id")
        keys.remove("cas")
        return [ResultWrapper(row["id"], row["cas"], row[keys[0]]) for row in rows_list]


import sys
import pytz
import base64
import re

from copy import deepcopy
import datetime
from fam.extra_types.lat_long import LatLong
from fractions import Fraction
from decimal import Decimal
from google.cloud.firestore_v1beta1 import GeoPoint


if sys.version_info[0] < 3:
    PYTHON_VERSION = 2
else:
    PYTHON_VERSION = 3


class BaseDataAdapter(object):

    # a fam doc to serialise into firestore
    def serialise(self, doc):
        dup = deepcopy(doc)
        return self._serialise_walk(dup)

    def deserialise(self, doc):
        dup = deepcopy(doc)
        return self._deserialise_walk(dup)


    def is_a_string(self, node):

        if PYTHON_VERSION == 3:
            return isinstance(node, str)
        else:
            return isinstance(node, basestring)


    def is_a_number(self, node):

        if PYTHON_VERSION == 3:
            return isinstance(node, int) or isinstance(node, float)
        else:
            return isinstance(node, int) or isinstance(node, float) or isinstance(node, long)


    def _serialise_walk(self, node):

        if isinstance(node, dict):
            for k, v in node.items():
                node[k] = self._serialise_walk(v)
            return node
        if isinstance(node, list):
            return [self._serialise_walk(v) for v in node]
        if isinstance(node, LatLong):
            return self.serialise_lat_long(node)
        if isinstance(node, Fraction):
            return self.serialise_fraction(node)
        if isinstance(node, Decimal):
            return self.serialise_decimal(node)
        if isinstance(node, datetime.datetime):
            return self.serialise_date_time(node)
        if self.is_a_string(node):
            return self.serialise_string(node)
        if isinstance(node, bytes):
            return self.serialise_bytes(node)
        if isinstance(node, bool):
            return self.serialise_bool(node)
        if self.is_a_number(node):
            return self.serialise_number(node)
        if hasattr(node, "to_json"):
            return self.serialise_object(node)
        if node is None:
            return None

        raise Exception("BaseAdapter can't serialise this value: %s" % str(node))


    def serialise_lat_long(self, lat_long):
        return "::latlong::%s,%s" % (lat_long.latitude, lat_long.longitude)

    def serialise_fraction(self, fraction):
        return "::fraction::%s/%s" % (fraction.numerator, fraction.denominator)

    def serialise_decimal(self, decimal):
        return "::decimal::%s" % str(decimal)

    def serialise_date_time(self, date_time):
        utc = pytz.utc
        if date_time.tzinfo is None:
            dt = date_time.replace(tzinfo=utc)
        else:
            dt = date_time.astimezone(utc)
        as_iso = dt.isoformat("T") + "Z"
        fixed = as_iso.replace("+00:00", "")
        return "::datetime::%s" % str(fixed)


    def serialise_bytes(self, btes):
        encoded = base64.b64encode(btes)
        result = "::bytes::%s" % encoded.decode("utf-8")
        return result

    def serialise_string(self, string):
        return string

    def serialise_bool(self, boolean):
        return boolean

    def serialise_number(self, number):
        return number

    def serialise_object(self, obj):
        return obj.to_json()

    def is_legacy_datetime(self, node):

        if not self.is_a_string(node):
            return False

        datepattern = r"""^\d{4}-(0[1-9]|1[0-2])-([0-2]\d|3[0-1])T([0-1][0-9]|2[0-3]):[0-5]\d:[0-5]\d([.]\d{1,8})?Z$"""
        pattern = re.compile(datepattern)
        return pattern.match(node)


    def _deserialise_walk(self, node):


        if isinstance(node, dict):
            for k, v in node.items():
                node[k] = self._deserialise_walk(v)
            return node
        if isinstance(node, list):
            return [self._deserialise_walk(v) for v in node]
        if self.is_a_string(node):
            if node.startswith("::fraction::"):
                stripped = node[len("::fraction::"):]
                num, denom = stripped.split("/")
                return Fraction(int(num), int(denom))
            if node.startswith("::decimal::"):
                stripped = node[len("::decimal::"):]
                return Decimal(stripped)
            if node.startswith("::latlong::"):
                stripped = node[len("::latlong::"):]
                lat, long = stripped.split(",")
                return LatLong(float(lat), float(long))
            if node.startswith("::bytes::"):
                stripped = node[len("::bytes::"):]
                return base64.b64decode(stripped)
            if node.startswith("::datetime::"):
                stripped = node[len("::datetime::"):]
                if "." in stripped:
                    dt = datetime.datetime.strptime(stripped, '%Y-%m-%dT%H:%M:%S.%fZ')
                else:
                    dt = datetime.datetime.strptime(stripped, '%Y-%m-%dT%H:%M:%SZ')
                dt = dt.replace(tzinfo=pytz.utc)
                return dt
            if self.is_legacy_datetime(node):
                stripped = node
                if "." in stripped:
                    dt = datetime.datetime.strptime(stripped, '%Y-%m-%dT%H:%M:%S.%fZ')
                else:
                    dt = datetime.datetime.strptime(stripped, '%Y-%m-%dT%H:%M:%SZ')
                dt = dt.replace(tzinfo=pytz.utc)
                return dt
            return node
        if isinstance(node, bool):
            return node
        if self.is_a_number(node):
            return node

        raise Exception("FirestoreDataAdapter can't deserialise this value: %s", node)


from __future__ import absolute_import
import os
import unittest
from fam.database import CouchDBWrapper
from fam.mapper import ClassMapper
from fam.buffer import buffered_db
from fam.tests.test_couchdb.config import *
from fam.tests.models.test01 import Dog, Cat, Person

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(THIS_DIR, "data")

class CacheTests(unittest.TestCase):

    def setUp(self):
        mapper = ClassMapper([Dog, Cat, Person])
        self.db = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True)
        self.db.update_designs()

    def tearDown(self):
        pass

    def test_delete_cat_refs(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        cat.save(self.db)
        key = cat.key
        cat2 = Cat(name="puss", owner_id=paul.key, legs=2)
        cat2.save(self.db)
        revivedcat1 = self.db.get(key)

        self.assertTrue(revivedcat1 is not None)

        paul.delete(self.db)
        revivedcat2 = self.db.get(key)
        print("revivedcat2:" , revivedcat2)
        self.assertTrue(revivedcat2 is None, "revivedcat2: %s" % revivedcat2)


import simplejson as json
import traceback
import uuid
import datetime
import types
import sys
import six
from copy import deepcopy

from fam.fam_json import object_default

from .constants import *
from .exceptions import *
from .fields import *

__all__ = [
    "BoolField",
    "NumberField",
    "StringField",
    "ListField",
    "DictField",
    "ObjectField",
    "LatLongField",
    "DateTimeField",
    "BytesField",
    "DecimalField",
    "FractionField",
    "ReferenceTo",
    "ReferenceFrom",
    "EmailField",
    "GenericObject",
    "FamObject"
]



def current_xml_time():
    now = datetime.datetime.utcnow()
    return now.strftime("%Y-%m-%dT%H:%M:%SZ")


## this provides a class based wrapper


class GenericMetaclass(type):

    def __new__(cls, name, bases, dct):
        attrs = dct.copy()
        # take a copy of this classes own fields
        attrs["cls_fields"] = {} if attrs["fields"] is None else attrs["fields"].copy()

        if attrs.get("fields") is None:
            attrs["fields"] = {}
        for b in bases:
            if hasattr(b, "fields"):
                attrs["fields"].update(b.fields)

        #check that the names of all ref to fields end with _id
        for fieldname, field in attrs["fields"].items():
            if isinstance(field, ReferenceTo) and not fieldname.endswith("_id"):
                raise FamError("All ReferenceTo field names must end with _id, %s doesn't" % fieldname)

        attrs[TYPE_STR] = name.lower()
        newcls = super(GenericMetaclass, cls).__new__(cls, name, bases, attrs)
        module = sys.modules[newcls.__module__]
        if hasattr(module, "NAMESPACE"):
            attrs[NAMESPACE_STR] = module.NAMESPACE
        else:
            attrs[NAMESPACE_STR] = "genericbase"
        newcls = super(GenericMetaclass, cls).__new__(cls, name, bases, attrs)
        return newcls


# class FamObject(object, metaclass=GenericMetaclass):
class FamObject(six.with_metaclass(GenericMetaclass)):
    use_rev = True
    additional_properties = False
    grants_access = False
    # extra sync gateway keywords that are used by sync function
    sg_allow_public_write = False

    # __metaclass__ = GenericMetaclass

    fields = {}
    acl = None

    def __init__(self, key=None, rev=None, **kwargs):

        type_name = self.__class__.__name__.lower()

        namespace = self.__class__.namespace.lower()

        self.key = key if key is not None else u"%s_%s" % (type_name, str(uuid.uuid4()))
        if rev is not None:
            self.rev = rev

        self._properties = {}

        for k, v in kwargs.items():
            if not k.startswith("_"):
                setattr(self, k, v)

        self._db = None

        if kwargs.get(TYPE_STR) is None:
            self._properties[TYPE_STR] = type_name

        self._check_defaults()

        if kwargs.get(NAMESPACE_STR) is None:
            self._properties[NAMESPACE_STR] = namespace
        else:
            if self._properties[NAMESPACE_STR] != namespace:
                raise Exception("the given namespace doesn't match the class")

    @classmethod
    def all(cls, db):
        return db.get_all_type(cls.namespace, cls.type)


    def _get_namespace(self):
        return self._properties.get(NAMESPACE_STR)

    def _get_type(self):
        return self._properties.get(TYPE_STR)

    def _get_properties(self):
        prop = self._properties.copy()
        del prop[TYPE_STR]
        del prop[NAMESPACE_STR]
        return prop


    def _check_defaults(self):
        for field_name, field in self.__class__.fields.items():
            if field.default is not None and not field_name in self._properties:
                self._properties[field_name] = field.get_default()

    def _check_immutable(self, existing):
        for field_name, field in self.__class__.fields.items():
            if field.immutable:
                attr = getattr(existing, field_name)
                if attr is not None and getattr(self, field_name) != attr:
                    raise FamImmutableError("You can't change the value of {} on a {} it has been made immutable: {}".format(field_name, self.__class__.__name__, getattr(existing, field_name)))





    @classmethod
    def get_unique_instance(cls, db, field_name, value):

        field = cls.fields.get(field_name)
        if field is None:
            return None

        if not field.unique:
            return None

        type_name = cls._type_with_ref(field_name)
        return db.get_unique_instance(cls.namespace, type_name, field_name, value)


    @classmethod
    def create(cls, db, key=None, **kwargs):
        obj = cls(key=key, **kwargs)
        obj._pre_save_new_cb(db)
        created = db.set_object(obj)
        if obj.use_rev and hasattr(created, "rev") and created.rev is not None:
            obj.rev = created.rev
        obj._post_save_new_cb(db)
        obj._db = db
        return obj


    def save_without_checks(self, db):
        self._pre_save_new_cb(db)
        created = db.set_object(self)
        if self.use_rev and hasattr(created, "rev") and created.rev is not None:
            self.rev = created.rev
        self._post_save_new_cb(db)
        self._db = db


    def save(self, db):

        if db.check_on_save:
            return self.save_with_checks(db)
        else:
            return self.save_without_checks(db)


    def save_with_checks(self, db):

        self._db = db

        existing = FamObject.get(db, self.key, class_name=self.type)

        if existing:
            # doc = result.value
            rev = existing.rev

            self._check_immutable(existing)

            # raise a conflict if revs different
            if self.use_rev and rev != self.rev:
                if not self.resolve_write_conflict(existing, rev):
                    raise FamResourceConflict("bad rev id: %s, rev: %s db_rev: %s" % (self.key, self.rev, rev))

            self._pre_save_update_cb(db, existing._properties)
            # just force the rev if not using it
            result = db.set_object(self, rev=self.rev if self.use_rev else rev)

            if self.use_rev and hasattr(result, "rev"):
                self.rev = result.rev

            self._post_save_update_cb(db)
            updated = True
        else:
            self._pre_save_new_cb(db)
            result = db.set_object(self)
            self._post_save_new_cb(db)
            updated = False

        if self.use_rev and hasattr(result, "rev"):
            self.rev = result.rev

        return updated


    def resolve_write_conflict(self, existing_doc, existing_rev):
        return False


    def delete(self, db):
        self._pre_delete_cb(db)
        db._delete(self.key, self.rev, self.type)
        self._post_delete_cb(db)
        self.delete_references(db)
        # db.sync_up()


    @classmethod
    def delete_key(cls, db, key):
        obj = cls.get(db, key)
        obj.delete(db)



    def delete_references(self, db):
        fields_set = list(self.__class__.fields.items())
        for field_name, field in fields_set:
            if isinstance(field, ReferenceTo):
                if field.cascade_delete:
                    if field_name.endswith("_id"):
                        obj = getattr(self, field_name[:-3])
                        if obj is not None:
                            obj.delete(db)
                    else:
                        raise Exception("should have _id")

            if isinstance(field, ReferenceFrom):
                type_name = self.__class__._type_with_ref(field_name)
                objs = self._db.get_refs_from(self.namespace, type_name, field_name, self.key, field)

                if field.cascade_delete:
                    for obj in objs:
                        obj.delete(db)
                else:
                    for obj in objs:
                        del obj._properties[field.fkey]
                        obj.save(db)

    def __str__(self):
        return self.as_json()


    def as_dict(self):
        d = {}
        d.update(self.properties)
        d[NAMESPACE_STR] = self.namespace
        d[TYPE_STR] = self.type
        d["_id"] = self.key
        if self.rev is not None:
            d["_rev"] = self.rev
        return d

    def as_json(self):
        d = {}
        d.update(self.properties)
        d[NAMESPACE_STR] = self.namespace
        d[TYPE_STR] = self.type
        d["_id"] = self.key
        if self.rev is not None:
            d["_rev"] = self.rev
        return json.dumps(d, sort_keys=True, indent=4, default=object_default, separators=(',', ': '))


    def __eq__(self, other):

        if self.key != other.key:
            return False

        if self.use_rev and self.rev != other.rev:
            return False

        if self.type != other.type:
            return False

        keys = set(other.properties.keys()).intersection(set(self.properties.keys()))

        for k in keys:
            if k != "schema" and k != "channels":
                if other._properties[k] != self._properties[k]:
                    return False
        return True


    @classmethod
    def get(cls, db, key, class_name=None):
        # ugly thing to get around cache double dispatch
        cn = class_name if class_name is not None else cls.__name__.lower()
        if not hasattr(db, "_get"):
            #this will call back here but using the correct db
            return db.get(key, class_name=cn)

        result = db._get(key, class_name=cn)
        if result is None:
            return None
        doc = result.value
        rev = result.rev
        return cls._from_doc(db, key, rev, doc)


    @classmethod
    def _from_doc(cls, db, key, rev, doc):

        if "_id" in doc.keys():
            del doc["_id"]

        if "_rev" in doc.keys():
            del doc["_rev"]

        correctCls = db.class_for_type_name(doc.get(TYPE_STR), doc.get(NAMESPACE_STR))
        if correctCls is None:
            raise Exception("couldn't find class {} for doc {}".format(doc.get(TYPE_STR), json.dumps(doc, indent=4)))

        obj = correctCls(key=key, rev=rev, **doc)
        obj._db = db
        return obj


    @classmethod
    def from_json(cls, db, as_json):
        key = as_json["key"]
        rev = as_json.get("rev")
        doc = as_json["properties"].copy()
        doc[NAMESPACE_STR] = as_json[NAMESPACE_STR]
        doc[TYPE_STR] = as_json[TYPE_STR]

        return cls._from_doc(db, key, rev, doc)


    def _pre_save_new_cb(self, db):
        if hasattr(self, "pre_save_new_cb"):
            self.pre_save_new_cb(db)
    
    def _post_save_new_cb(self, db):
        if hasattr(self, "post_save_new_cb"):
            self.post_save_new_cb(db)

    def _changes_cb(self, db, queue, new=False, **kwargs):
        if hasattr(self, "changes_cb"):
            self.changes_cb(db, queue, new=new, **kwargs)

    def _pre_save_update_cb(self, db, old_properties):
        if hasattr(self, "pre_save_update_cb"):
            self.pre_save_update_cb(db, old_properties)

    def _post_save_update_cb(self, db):
        if hasattr(self, "post_save_update_cb"):
            self.post_save_update_cb(db)

    def _pre_delete_cb(self, db):
        if hasattr(self, "pre_delete_cb"):
            self.pre_delete_cb(db)

    def _post_delete_cb(self, db):
        if hasattr(self, "post_delete_cb"):
            self.post_delete_cb(db)


    @classmethod
    def n1ql(cls, db, query, with_revs=False, *args, **kwargs):
        if with_revs:
            rows = db._n1ql_with_rev(query, *args, **kwargs)
        else:
            rows = db._n1ql(query, *args, **kwargs)
        return [GenericObject._from_doc(db, row.key, None, row.value) for row in rows]



    @classmethod
    def view(cls, db, view_name, **kwargs):
        if db.database_type in ["sync_gateway", "null"]:
            rows = db.view(view_name, **kwargs)
            return [GenericObject._from_doc(db, row.key, row.rev, row.value) for row in rows]
        else:
            return cls.view_iterator(db, view_name, **kwargs)


    @classmethod
    def view_iterator(cls, db, view_name, **kwargs):

        if "limit" in kwargs:
            rows = db.view(view_name, **kwargs)
            for row in rows:
                yield GenericObject._from_doc(db, row.key, row.rev, row.value)
        else:

            limit = 100
            skip = 0

            while True:
                rows =  db.view(view_name, limit=limit, skip=skip, **kwargs)
                if len(rows) == 0:
                    break
                skip += limit
                for row in rows:
                    yield GenericObject._from_doc(db, row.key, row.rev, row.value)



    @classmethod
    def changes(cls, db, **kwargs):
        last_seq, rows = db._changes(**kwargs)

        changeset = []

        for row in rows:
            try:
                changeset.append(GenericObject._from_doc(db, row.key, row.rev, row.value))
            except Exception as e:
                print("BAD!!! swallowing all exceptions in blud changes")


        return last_seq, changeset
        # return last_seq, [GenericObject._from_doc(db, row.key, row.rev, row.value) for row in rows]

    @classmethod
    def _query_view(cls, db, view_name, key):
        rows =  db.view(view_name, key=key)
        return [GenericObject._from_doc(db, row.key, row.rev, row.value) for row in rows]


    @classmethod
    def from_row(cls, db, row):
        return GenericObject._from_doc(db, row.key, row.rev, row.value)


    @classmethod
    def _type_with_ref(cls, name):

        if name in cls.cls_fields.keys():
            return cls.type

        for base in cls.__bases__:
            found = base._type_with_ref(name)
            if found:
                return found

        return None



    def __getattr__(self, name):
        field = self.__class__.fields.get(name)
        if isinstance(field, ReferenceFrom):
            if self._db is None:
                traceback.print_stack()
                raise Exception("no db")
            # look at super class gypes to find clas with ref
            type_name = self.__class__._type_with_ref(name)
            return self._db.get_refs_from(self.namespace, type_name, name, self.key, field)


        if "%s_id" % name in self.properties.keys():
            id_name = "%s_id" % name
            ref = self.__class__.fields.get(id_name)
            if isinstance(ref, ReferenceTo):
                if self._db is None:
                    traceback.print_stack()
                    raise Exception("no db")
                # return GenericObject.get(self._db, self._properties[id_name])
                return self._db.get(self._properties[id_name], class_name=ref.refcls)

        if name in self._properties.keys():
            # print("found in properties", name)
            # if it is a subclass of stringfield for string formats
            # if isinstance(field, StringField) and not field.__class__ == StringField:
            #     if hasattr(field, "from_json"):
            #         return field.from_json(self._properties[name])
            return self._properties[name]


        if name in self.__class__.fields:
            return None
        elif "%s_id" % name in self.__class__.fields:
            return None
        elif name == "rev":
            return None
        else:
            # print("self: ", self, name)
            raise AttributeError("Not found %s" % name)


    def update(self, values):
        if "_db" in self.__dict__:
            if hasattr(self._db, "update"):
                self._db.update(self.namespace, self.type, self.key, values)

        for k, v in values.items():
            setattr(self, k, v)


    def _update_property(self, key, value, field):

        if value is None:
            if key in self._properties:
                del self._properties[key]
        else:
            self._properties[key] = value


    def __setattr__(self, name, value):

        if name in RESERVED_PROPERTY_NAMES:
            self.__dict__[name] = value
            return

        field = self.fields.get(name)

        alias = "%s_id" % name
        if alias in self.fields.keys():
            self._update_property(alias, value.key, field)

        elif name in self.fields or name in ("type", "namespace", "schema", "update_seconds", "update_nanos") or self.additional_properties :
            if field is not None:
                if field.immutable and name in self._properties:
                    raise FamImmutableError("You cannot change the immutable property %s" % name)
                if isinstance(field, ObjectField) and not isinstance(value, field.cls):
                    if hasattr(field.cls, "from_json"):
                        value = field.cls.from_json(value)
                    else:
                        value = field.from_json(value)

                if isinstance(field, ListField) and field.item_cls is not None:
                    #cast the items in the list into a certain class
                    value = [field.item_cls.from_json(i) for i in value]

                # if issubclass(field.__class__, StringField) and not (isinstance(value, str) or value is None):
                #     value = field.to_json(value)

            self._update_property(name, value, field)
        elif name.startswith("_"):
            self.__dict__[name] = value
            return
        else:
            raise FamValidationError("""You cant use the property name %s on the class %s
            If you would like to set additional properties on this class that are not specified
            set the class attribute "additional_properties" to True.
            """ % (name, self.__class__.__name__))


    namespace = property(_get_namespace)
    type = property(_get_type)
    properties = property(_get_properties)


GenericObject = FamObject



from .basic_tests import BasicBaseTestCases
from .anything_tests import AnythingBaseTests
from .callback_tests import CallbackBaseTests
from .field_attribute_tests import FieldAttributeBaseTests
from .schema_tests import SchemaBaseTests
from .index_tests import IndexBaseTests

common_test_classes = [BasicBaseTestCases.BasicTests,
                        BasicBaseTestCases.RefNameTests,
                        AnythingBaseTests.AnythingTests,
                        CallbackBaseTests.CallbackTests,
                        FieldAttributeBaseTests.FieldAttributeTests,
                        SchemaBaseTests.SchemaTests,
                        IndexBaseTests.IndexTests
                        ]


import unittest
import os
import time
import json
import subprocess
from fam.database import SyncGatewayWrapper
from fam.mapper import ClassMapper
from fam.tests.test_sync_gateway.config import *
from fam.tests.models.acl import Car, Bike, Boat
from fam.acl.writer import write_sync_function, _requirements_from_mapper
from fam.utils import couchbase_utils

from fam.exceptions import *

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(TEST_DIR, "data")

"""
Actors are people who do things

users: a set of named users
roles: a set of people who have any of given roles
owner: an owner as named in the doc
anyone: anyone as long as we know who they are

Actions are things they can do:

create: Create a new document with a new id
update: Make a change to an existing document
delete: Delete an existing document

"""


class testPermissions(unittest.TestCase):

        def setUp(self):

            self.gateway = None
            self.db = None

            self.mapper = ClassMapper([Car, Boat])

            self.start_gateway()



        def start_gateway(self):

            cmd = "{} -log=* -url walrus: ".format(SYNC_GATEWAY_PATH)
            # print cmd

            time.sleep(0.25)
            self.gateway = subprocess.Popen(cmd, shell=True)
            time.sleep(0.25)

            admin_url = "http://%s:%s" % (SYNC_GATEWAY_ADMIN_HOST, SYNC_GATEWAY_ADMIN_PORT)
            self.admin_db = SyncGatewayWrapper(self.mapper, admin_url, SYNC_GATEWAY_NAME)
            self.admin_db.update_designs()
            self.add_users()


        def add_users(self):

            admin_url = "http://%s:%s" % (SYNC_GATEWAY_ADMIN_HOST, SYNC_GATEWAY_ADMIN_PORT)

            couchbase_utils.add_person_to_gateway(admin_url,
                                                  SYNC_GATEWAY_NAME,
                                                  "paul_id",
                                                  "paul",
                                                  "password1",
                                                  admin_channels=["cars", "paul"])

            couchbase_utils.add_person_to_gateway(admin_url,
                                                  SYNC_GATEWAY_NAME,
                                                  "sol_id",
                                                  "sol",
                                                  "password2",
                                                  admin_channels=["sol"])

            paul_url = "http://paul:password1@%s:%s" % (SYNC_GATEWAY_HOST, SYNC_GATEWAY_PORT)
            self.paul_db = SyncGatewayWrapper(self.mapper, paul_url, SYNC_GATEWAY_NAME)

            sol_url = "http://sol:password2@%s:%s" % (SYNC_GATEWAY_HOST, SYNC_GATEWAY_PORT)
            self.sol_db = SyncGatewayWrapper(self.mapper, sol_url, SYNC_GATEWAY_NAME)


        def tearDown(self):
            # stop the gateway
            if self.gateway is not None:
                self.gateway.kill()


        def test_get_user(self):

            user_info = self.admin_db.user("paul")
            # print user_info
            self.assertTrue(user_info != None)
            roles = user_info["admin_roles"]

            self.assertEqual(roles, ["paul_id"])


        def test_create_role(self):

            role_info = self.admin_db.role("new_role")
            # print "role: ",  role_info
            self.assertTrue(role_info == None)

            self.admin_db.ensure_role("new_role")

            role_info = self.admin_db.role("new_role")
            # print "role: ",  role_info
            self.assertTrue(role_info != None)


        def test_add_role(self):

            user_info = self.admin_db.user("paul")
            self.assertTrue(user_info != None)
            roles = user_info["admin_roles"]
            self.assertEqual(roles, ["paul_id"])
            channels = user_info["admin_channels"]
            self.assertEqual(set(channels), set(["cars", "paul"]))

            success = self.admin_db.ensure_user_role("paul", "new_role")

            self.assertTrue(success)

            user_info = self.admin_db.user("paul")
            self.assertTrue(user_info != None)
            roles = user_info["admin_roles"]
            channels = user_info["admin_channels"]
            self.assertEqual(set(roles), set(["paul_id", "new_role"]))
            self.assertEqual(set(channels), set(["cars", "paul"]))








import jsonschema
import os
import json
import inspect
import copy
from fam.blud import GenericObject
from fam.constants import *



from .writer import createJsonSchema


class ModelValidator(object):

    def __init__(self, classes=None, modules=None, schema_dir=None):

        self.reference_store = {}
        self.ref_schemas = {}
        self.schema_dir = schema_dir

        if classes is not None:
            self._add_classes(classes)

        if modules is not None:
            self._add_modules(modules)


    def iter_schemas(self):

        for k, schema in self.ref_schemas.items():
            namespace = k[0]
            typename = k[1]
            yield namespace, typename, schema


    def _add_classes(self, classes):
        for cls in classes:
            type_name = cls.__name__.lower()
            namespace = cls.namespace.lower()
            self.add_schema(namespace, type_name, cls)


    def _add_modules(self, modules):
        for module in modules:
            classes = []
            for k, obj in module.__dict__.items():
                if inspect.isclass(obj):
                    if issubclass(obj, GenericObject):
                        if obj != GenericObject:
                            if not k.startswith("_"):
                                classes.append(obj)
            self._add_classes(classes)


    def add_schema(self, namespace, type_name, cls):

        schema = createJsonSchema(cls)

        if self.schema_dir is not None:
            schema["id"] = self._check_for_changes(namespace, type_name, schema)
        else:
            schema["id"] = "%s/%s" % (namespace, type_name)

        jsonschema.Draft4Validator.check_schema(schema)
        self.reference_store[schema["id"]] = schema
        self.ref_schemas[(namespace, type_name)] = schema


    def schema_id_for(self, namespace, type_name):
        schema = self.ref_schemas.get((namespace, type_name))
        if schema:
            # print schema["id"]
            return schema["id"]
        return None


    def validate(self, doc):

        schema_id = doc.get("schema")
        if schema_id is None:
            namespace = doc.get("namespace")
            type_name = doc.get("type")

            if namespace and type_name:
                schema_id = self.schema_id_for(namespace, type_name)

        if schema_id:
            schema = self._look_schema_with_lazy_load(schema_id)
            resolver = jsonschema.RefResolver(schema_id, schema, store=self.reference_store)
            validator = jsonschema.Draft4Validator(schema, resolver=resolver)
            validator.validate(doc)


    def _look_schema_with_lazy_load(self, schema_id):
        schema = self.reference_store.get(schema_id)
        if schema is None:
            schema = self._schema_from_id(schema_id)
            self.reference_store[schema_id] = schema

        return schema

        return ''.join(diff)

    def _schemata_are_equal(self, schema_a, schema_b):


        if schema_b is None:
            return False

        # expects dicts
        # don't mess with the input
        dupe_a = copy.deepcopy(schema_a)
        dupe_b = copy.deepcopy(schema_b)

        # remove ids before comparison
        if "id" in dupe_a:
            del dupe_a["id"]

        if "id" in dupe_b:
            del dupe_b["id"]

        return dupe_a == dupe_b


    def _type_dir(self, namespace, type_name):
        ns = namespace.replace("/", "_")
        dir_path = os.path.join(self.schema_dir, "schemata", ns, type_name)
        return dir_path

    def _schema_path(self, namespace, type_name, timestamp):
        type_dir = self._type_dir(namespace, type_name)
        dir_path = os.path.join(type_dir, timestamp)
        return dir_path


    def _timestamp_from_schema_id(self, schema_id):
        namespace, typename, timestamp = self._namespace_typename_timestamp_from_schema_id(schema_id)
        return timestamp


    def _schema_path_from_id(self, schema_id):
        namespace, type_name, timestamp = self._namespace_typename_timestamp_from_schema_id(schema_id)
        return self._schema_path(namespace, type_name, timestamp)


    def _schema_from_id(self, schema_id):
        schema_path = self._schema_path_from_id(schema_id)
        return self.schema_at_schema_path(schema_path)


    def _namespace_typename_timestamp_from_schema_id(self, schema_id):

        parts = schema_id.split("/")
        namespace = "/".join(parts[:-3])
        type_name = parts[-3]
        timestamp = parts[-2]

        return namespace, type_name, timestamp

    def schema_at_schema_path(self, schema_path):
        schema_path = os.path.join(schema_path, "schema.json")
        with open(schema_path, "r") as f:
            return json.loads(f.read())


    def _previous_schema(self, namespace, type_name):

        dir_path = self._type_dir(namespace, type_name)

        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        filenames = os.listdir(dir_path)
        if not filenames:
            return None

        most_recent_filename = sorted(filenames)[-1]
        return self.schema_at_schema_path(os.path.join(dir_path, most_recent_filename))

    def _check_for_changes(self, namespace, type_name, schema):

        existing_schema = self._previous_schema(namespace, type_name)

        if existing_schema is None:
            raise NotImplementedError("The schema for %s %s is missing" % (namespace, type_name))
        else:
            # compare the latest schema with the most recent
            if self._schemata_are_equal(schema, existing_schema):
                return existing_schema["id"]
            else:
                raise NotImplementedError("The schema for %s %s is not up to date" % (namespace, type_name))






import unittest
from fam.exceptions import *
from fam.tests.models.test01 import Cat, Person, Monkey


class FieldAttributeBaseTests:

    class FieldAttributeTests(unittest.TestCase):

        db = None

        def make_red(self, animal):
            animal.colour = "red"

        def cut_off_tail(self, animal):
            animal.tail = False

        def get_address(self, animal):
            return animal.address

        def test_immutability(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(colour="tabby", legs=4, owner=paul)
            self.assertRaises(FamImmutableError, self.make_red, cat)

        def test_immutability_on_non_existant_value(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(legs=4, owner=paul)
            self.make_red(cat)

        def test_getting_absent_no_defaults_returns_none(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(legs=4, owner=paul)
            name = cat.name
            self.assertEqual(name, None)

        def test_getting_unknown_fails(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(legs=4, owner=paul)
            self.assertRaises(AttributeError, self.get_address, cat)

        def test_immutable_defaults(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(legs=4, owner=paul)
            self.assertTrue(cat.tail == True)
            self.assertRaises(FamImmutableError, self.cut_off_tail, cat)

        def test_immutable_saving(self):

            monkey = Monkey(key="red_monkey", colour="red")
            self.db.put(monkey)
            monkey = Monkey(key="red_monkey", colour="blue")
            self.assertRaises(FamImmutableError, self.db.put, monkey)


        def test_using_default_and_required_fails(self):

            def duff_import():
                from fam.tests.models import test03
                print(test03)

            self.assertRaises(FamError, duff_import)


from fam.blud import GenericObject, StringField, ReferenceFrom, ReferenceTo, BoolField, NumberField, DictField


NAMESPACE = "glowinthedark.co.uk/test/2"


class Dog(GenericObject):
    additional_properties = True
    fields = {
        "name": StringField(),
        "owner": ReferenceTo(NAMESPACE, "person", cascade_delete=True)
        }

    def talk(self):
        return "woof"


class Person(GenericObject):
    fields = {
        "name": StringField(),
        "dogs": ReferenceFrom(NAMESPACE, "dog", "owner_id")
        }




"""
This either imports all of requests or wraps google app engine's urlfetch with requests like syntax

It is far from complete and just does the things I needed at the time
"""


try:
    from requests import *
except ImportError as e:
    import urllib
    import json
    from google.appengine.api import urlfetch
    
    class ResponseWrapper(object):
        
        def __init__(self, response):
            self.response = response
            
        @property
        def status_code(self):
            return int(self.response.status_code)
        
        @property
        def headers(self):
            return self.response.headers
        
        @property
        def content(self):
            return self.response.content

        def json(self):
            return json.loads(self.response.content)
        
        
    def get(url, **kwargs):
        return request("GET", url, **kwargs)

    def put(url, **kwargs):
        return request("PUT", url, **kwargs)
    
    def post(url, **kwargs):
        return request("POST", url, **kwargs)
    
    
    def request(method, url, **kwargs):
        
        if "params" in kwargs.keys():
            querystring = urllib.urlencode(kwargs["params"])
            url = "%s?%s" % (url, querystring)
        
        if "data" in kwargs.keys():
            data = kwargs["data"]
            if type(data) == type({}):
                payload = urllib.urlencode(data)
            elif type(data) == type(""):
                payload = data
            elif hasattr(data, "read"):
                payload = data.read()
            else:
                payload = None
        else:
            payload = None
        
        if "headers" in kwargs.keys():
            headers = kwargs["headers"]
        else:
            headers = {}
            
        if "allow_redirects" in kwargs.keys():
            follow_redirects = kwargs["allow_redirects"]
        else:
            follow_redirects = True

        if "timeout" in kwargs.keys():
            deadline = kwargs["timeout"]
        else:
            deadline = 5
            
        if "verify" in kwargs.keys():
            validate_certificate = kwargs["validate_certificate"]
        else:
            validate_certificate = False
        
        resp = urlfetch.fetch(url, 
                              payload=payload, 
                              method=method, 
                              headers=headers, 
                              allow_truncated=False, 
                              follow_redirects=follow_redirects, 
                              deadline=deadline, 
                              validate_certificate=validate_certificate)
        
        return ResponseWrapper(resp)



import unittest
import json
import time
import os

import firebase_admin
from firebase_admin import credentials
from firebase_admin import firestore

import fam
from fam.exceptions import *
from fam.tests.models.test01 import GenericObject, Dog, Cat, Person, JackRussell, Monkey, Monarch, NAMESPACE

from fam.database import FirestoreWrapper
from fam.mapper import ClassMapper

SECRETS_DIR = os.path.join(os.path.dirname(fam.__file__), "tests", "secrets")


class TestDB(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        creds_path = os.path.join(SECRETS_DIR, "earth-rover-test-d241bce5266d.json")
        mapper = ClassMapper([Dog, Cat, Person, JackRussell, Monkey, Monarch])
        cls.db = FirestoreWrapper(mapper, creds_path, namespace=NAMESPACE)
        cls.clear_db()


    @classmethod
    def clear_db(cls):
        cls.db.delete_all("dog")
        cls.db.delete_all("cat")
        cls.db.delete_all("person")
        cls.db.delete_all("jackrussell")
        cls.db.delete_all("monkey")
        cls.db.delete_all("dog__kennel_club_membership")


    def test_update_cat(self):
        paul = Person(name="paul")
        paul.save(self.db)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        cat.save(self.db)
        cat.name = "blackie"
        cat.save(self.db)

        self.assertEqual(cat.name, "blackie")
        self.assertEqual(cat._properties["name"], "blackie")
        self.assertFalse("name" in cat.__dict__.keys())

        got = Cat.get(self.db, cat.key)

        self.assertEqual(got.name, "blackie")





from fam.blud import GenericObject, StringField, ReferenceFrom, ReferenceTo, BoolField, NumberField, DictField, ObjectField, ListField
from fam.acl import CreateRequirement, DeleteRequirement, UpdateRequirement, NO_ONE, ANYONE

NAMESPACE = "glowinthedark.co.uk/test"



class Bike(GenericObject):

    fields = {
        "wheels": NumberField(),
        }


class Car(GenericObject):

    fields = {
        "colour": StringField(),
        "stars": NumberField(),
        "owner_name": StringField(),
        "channels": ListField(),
        "access":  ListField()
        }

    grants_access = True

    acl = [
        CreateRequirement(role=None, owner=True),
        DeleteRequirement(role=None, owner=True),
        UpdateRequirement(role=[], fields=["access"]),
        UpdateRequirement(role=None, owner=True, fields=["colour"]),
    ]

class Boat(GenericObject):

    fields = {
        "name": StringField(),
        "is_sail": BoolField(),
        "owner_name": StringField(),
        "channels": ListField(),
        "access": ListField()
    }

    grants_access = True

    acl = [
        CreateRequirement(role=ANYONE, owner=True, access=False),
        UpdateRequirement(role=NO_ONE),
        DeleteRequirement(role=ANYONE, owner=True),
    ]





import json
import time
import sys
import copy

import requests
from requests.exceptions import HTTPError

import firebase_admin
from firebase_admin import credentials
from firebase_admin import firestore
import jsonschema

from google.cloud.exceptions import NotFound

from fam.exceptions import *
from fam.constants import *
from fam.database.base import BaseDatabase
from fam.database.couchdb import ResultWrapper
from fam.database.firestore_adapter import FirestoreDataAdapter

from grpc._channel import _Rendezvous

from fam.blud import GenericObject

from .custom_token import CustomToken

if sys.version_info[0] < 3:
    PYTHON_VERSION = 2
else:
    PYTHON_VERSION = 3


"""

get(key, class_name)
update(key, class_name, **kwargs)
delete(item)


"""

def refresh_check(func):
    def func_wrapper(instance, *args, **kwargs):
        now = time.time()
        if instance.expires is not None and instance.expires < now + 10:
            instance.refresh()
        try:
            return func(instance, *args, **kwargs)
        except _Rendezvous as e:
            if instance.expires is not None:
                instance.refresh()
            return func(instance, *args, **kwargs)
    return func_wrapper


def raise_detailed_error(request_object):
    try:
        request_object.raise_for_status()
    except HTTPError as e:
        # raise detailed error message
        # TODO: Check if we get a { "error" : "Permission denied." } and handle automatically
        raise HTTPError(e, request_object.text)



class FirestoreWrapper(BaseDatabase):

    def query_view(self, view_name, **kwargs):
        raise NotImplementedError

    def changes(self, since=None, channels=None, limit=None, feed=None, timeout=None, filter=None):
        raise NotImplementedError

    database_type = "firestore"
    check_on_save = False

    def __init__(self,
                 mapper,
                 creds_path,
                 project_id=None,
                 custom_token=None,
                 api_key=None,
                 validator=None,
                 read_only=False,
                 name=None,
                 namespace=None
                 ):

        self.mapper = mapper
        self.validator = validator
        self.read_only = read_only
        self.api_key = api_key
        self.namespace = namespace

        self.expires = None

        self.data_adapter = FirestoreDataAdapter()

        # Use a service account

        options = {"httpTimeout": 5}

        app = None

        if custom_token is not None:
            # from device client
            self.user = self.sign_in_with_custom_token(custom_token)
            self.update_expires()
            self.creds = CustomToken(self.user["idToken"], project_id)

            app_name = name if name else firebase_admin._DEFAULT_APP_NAME

            try:
                app = firebase_admin.get_app(name=app_name)
            except ValueError as e:
                # ugly hack to account for different behaviour of google libs
                try:
                    app = firebase_admin.initialize_app(self.creds, name=app_name, options=options)
                except Exception as e:
                    self.creds = self.creds.get_credential()
                    app = firebase_admin.initialize_app(self.creds, name=app_name, options=options)


        elif creds_path is not None:
            # in dev with service creds
            try:
                self.creds = credentials.Certificate(creds_path)
                app = firebase_admin.initialize_app(self.creds, options=options)
            except ValueError as e:
                print("already initaialised")
        else:
            # in app engine environment
            already_initialised = firebase_admin._DEFAULT_APP_NAME in firebase_admin._apps
            if not already_initialised:
                app = firebase_admin.initialize_app(options=options)

        self.db = firestore.client(app=app)


    def update_expires(self):
        self.expires = time.time() + int(self.user["expiresIn"])


    def sign_in_with_custom_token(self, token):
        request_ref = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/verifyCustomToken?key={0}".format(self.api_key)
        headers = {"content-type": "application/json; charset=UTF-8"}
        data = json.dumps({"returnSecureToken": True, "token": token})
        request_object = requests.post(request_ref, headers=headers, data=data)
        raise_detailed_error(request_object)
        return request_object.json()


    def refresh(self):

        refresh_token = self.user["refreshToken"]
        request_ref = "https://securetoken.googleapis.com/v1/token?key={0}".format(self.api_key)
        headers = {"content-type": "application/json; charset=UTF-8"}
        data = json.dumps({"grantType": "refresh_token", "refreshToken": refresh_token})
        request_object = requests.post(request_ref, headers=headers, data=data)
        raise_detailed_error(request_object)
        request_object_json = request_object.json()

        self.user = {
            "userId": request_object_json["user_id"],
            "idToken": request_object_json["id_token"],
            "refreshToken": request_object_json["refresh_token"],
            "expiresIn": request_object_json["expires_in"]
        }

        oauth_creds = self.creds.get_credential()
        oauth_creds.token = request_object_json["id_token"]


    def set_object(self, obj, rev=None):
        return self._set(obj.key, obj._properties)


    @refresh_check
    def _set(self, key, input_value, rev=None):

        value = self.data_adapter.serialise(input_value)
        self._check_uniqueness(key, value)

        if self.read_only:
            raise Exception("This db is read only")

        if self.validator is not None:
            if "namespace" in value and not "schema" in value:
                schema_id = self.validator.schema_id_for(value["namespace"], value["type"])
                # print("schema_id", schema_id)
                if schema_id is not None:
                    value["schema"] = schema_id
            try:
                self.validator.validate(value)
            except jsonschema.ValidationError as e:
                raise FamValidationError(e)

        type = value["type"]
        value["_id"] = key
        sans_metadata = copy.deepcopy(value)

        del sans_metadata["type"]
        del sans_metadata["namespace"]

        self.db.collection(type).document(key).set(sans_metadata)
        return ResultWrapper.from_couchdb_json(value)



    def _work_out_class(self, key, class_name):

        if isinstance(class_name, list):
            for cn in class_name:
                if key.startswith(cn):
                    return cn
            raise Exception("can't work out class name %s %s" % (key, class_name))
        else:
            return class_name


    @refresh_check
    def _get(self, key, class_name):

        single_class_name = self._work_out_class(key, class_name)
        doc_ref = self.db.collection(single_class_name).document(key)

        try:
            as_json = self.value_from_snapshot(doc_ref.get())
            return ResultWrapper.from_couchdb_json(as_json)
        except NotFound:
            return None


    def value_from_snapshot(self, snapshot):
        as_json = self.data_adapter.deserialise(snapshot.to_dict())
        # as_json["_id"] = snapshot.reference.id
        as_json["type"] = snapshot.reference.parent.id
        as_json["namespace"] = self.namespace
        return as_json


    def _get_refs_from(self, key, type_name, field_name):
        type_ref = self.db.collection(type_name)
        query_ref = type_ref.where(field_name, u'==', key)
        snapshots = query_ref.get()
        rows = [ResultWrapper.from_couchdb_json(self.value_from_snapshot(snapshot)) for snapshot in snapshots]
        objs = [GenericObject._from_doc(self, row.key, row.rev, row.value) for row in rows]
        return objs


    @refresh_check
    def get_all_type(self, namespace, type_name):
        all_sub_class_names = self.mapper.get_all_subclass_names(namespace, type_name)

        objs = []
        for type_name in all_sub_class_names:
            objs += self.get_single_type(namespace, type_name)
        return objs


    @refresh_check
    def get_single_type(self, namespace, type_name):
        type_ref = self.db.collection(type_name)
        docs = type_ref.get()
        rows = [ResultWrapper.from_couchdb_json(doc.to_dict()) for doc in docs]
        objs = [GenericObject._from_doc(self, row.key, row.rev, row.value) for row in rows]
        return objs


    @refresh_check
    def get_refs_from(self, namespace, type_name, name, key, field):
        all_sub_class_names = self.mapper.get_all_subclass_names(namespace, field.refcls)
        objs = []
        for class_name in all_sub_class_names:
            objs += self._get_refs_from(key, class_name, field.fkey)
        return objs


    @refresh_check
    def get_with_value(self, namespace, type_name, field_name, value):
        return self._get_refs_from(value, type_name, field_name)


    @refresh_check
    def _delete(self, key, rev, type_name):

        if self.read_only:
            raise Exception("This db is read only")
        self._clear_uniqueness_typed(key, type_name)
        self.db.collection(type_name).document(key).delete()


    @refresh_check
    def delete_all(self, type_name):
        coll_ref = self.db.collection(type_name)
        self._delete_collection(coll_ref, 10)

    def _query_items_simple(self, firebase_query):
        snapshots = firebase_query.get()
        results = []
        for snapshot in snapshots:
            wrapper = ResultWrapper.from_couchdb_json(self.value_from_snapshot(snapshot))
            results.append(GenericObject.from_row(self, wrapper))
        return results


    def query_items(self, firebase_query, batch_size=None, order_by=u'_id'):
        if batch_size is not None:
            return self.query_items_iterator(firebase_query, batch_size=batch_size, order_by=order_by)
        else:
            return self._query_items_simple(firebase_query)


    def query_snapshots(self, firebase_query, batch_size=100):
        return self.query_snapshots_iterator(firebase_query, batch_size=batch_size)


    def query_items_iterator(self, firebase_query, batch_size, order_by=u'_id'):

        for snapshot in self.query_snapshots_iterator(firebase_query, batch_size=batch_size, order_by=order_by):
            wrapper = ResultWrapper.from_couchdb_json(self.value_from_snapshot(snapshot))
            obj = GenericObject.from_row(self, wrapper)
            yield obj


    def query_snapshots_iterator(self, firebase_query, batch_size, order_by=u'_id'):

        skip = 0
        query = firebase_query.order_by(order_by).limit(batch_size)

        while True:
            docs = query.get()
            docs_list = list(docs)
            if len(docs_list) == 0:
                break
            for doc_snapshot in docs_list:
                yield doc_snapshot
            last_doc = docs_list[-1]
            last_value = last_doc.to_dict()[order_by]
            query = firebase_query.order_by(order_by).start_after({
                order_by: last_value
            }).limit(batch_size)


    @refresh_check
    def update(self, namespace, type_name, key, input_value):

        values = self.data_adapter.serialise(input_value)

        self._check_uniqueness_typed(namespace, type_name, key, values)
        doc_ref = self.db.collection(type_name).document(key)
        if self.read_only:
            raise Exception("This db is read only")

        for k, v in values.items():
            if v is None:
                values[k] = firestore.DELETE_FIELD
        doc_ref.update(values)


    def _delete_collection(self, coll_ref, batch_size):
        docs = coll_ref.limit(10).get()
        deleted = 0

        for doc in docs:
            doc.reference.delete()
            deleted = deleted + 1

        if deleted >= batch_size:
            return self._delete_collection(coll_ref, batch_size)


    ##############   unique stuff ##########################


    def _get_unique_ref(self, type_name, key, field_name, field_value):
        unique_type_name = "%s__%s" % (type_name, field_name)
        unique_doc_ref = self.db.collection(unique_type_name).document(field_value)
        try:
            unique_doc = unique_doc_ref.get()
            if unique_doc is None:
                return unique_doc_ref, unique_type_name, field_name
            ## if it exists then check to see if its owned by another
            if unique_doc.to_dict()["owner"] != key:
                raise FamUniqueError("The value %s for %s is already taken" % (field_value, field_name))
            else:
                # no op in the case where the value is already set
                return None
        except NotFound:
            ## go ahead and set the new one
            return unique_doc_ref, unique_type_name, field_name


    def _check_uniqueness(self, key, value):

        type_name = value["type"]
        namespace = value["namespace"]
        self._check_uniqueness_typed(namespace, type_name, key, value)


    def _clear_uniqueness_typed(self, key, type_name):

        doc_ref = self.db.collection(type_name).document(key)
        doc = doc_ref.get()
        as_dict = doc.to_dict()

        type_name = doc.reference.parent.id

        cls = self.mapper.get_class(type_name, self.namespace)

        for field_name, field_value in as_dict.items():
            if field_name in cls.fields:
                field = cls.fields[field_name]
                if field.unique:
                    unique_type_name = "%s__%s" % (type_name, field_name)
                    unique_doc_ref = self.db.collection(unique_type_name).document(field_value)
                    unique_doc_ref.delete()


    def _check_uniqueness_typed(self, namespace, type_name, key, value):

        cls = self.mapper.get_class(type_name, namespace)

        to_set = []

        # check all the unique objects and fail if any are bad before setting anything
        for field_name, field_value in value.items():
            if field_name in cls.fields:
                field = cls.fields[field_name]
                if field.unique:
                    ref_name_fieldname = self._get_unique_ref(type_name, key, field_name, field_value)
                    if ref_name_fieldname is not None:
                        to_set.append(ref_name_fieldname)

        if len(to_set) > 0:

            try:
                doc_ref = self.db.collection(type_name).document(key)
                doc = doc_ref.get()
                as_dict = doc.to_dict()
            except NotFound:
                as_dict = None

            for unique_doc_ref, unique_type_name, field_name in to_set:

                unique_doc_ref.set({"owner": key, "type_name": type_name})

                if as_dict is not None:
                    existing_key = as_dict.get(field_name)
                    if existing_key is not None:
                        existing_unique_doc_ref = self.db.collection(unique_type_name).document(existing_key)
                        existing_unique_doc_ref.delete()




    def set_unique_doc(self, type_name, key, field_name, value):
        doc_ref = self.db.collection(type_name).document(key)
        unique_type_name = "%s__%s" % (type_name, field_name)
        if value is not None:
            unique_doc_ref = self.db.collection(unique_type_name).document(value)
            try:
                unique_doc = unique_doc_ref.get()
                ## if it exists then check to see if its owned by another
                if unique_doc.to_dict()["owner"] != key:
                    raise FamUniqueError("The value %s for %s is already taken" % (value, field_name))
                else:
                    # no op in the case where the value is already set
                    return
            except NotFound:
                unique_doc_ref.set({"owner": key, "type_name": type_name})

        # delete any existing
        try:
            doc = doc_ref.get()
            as_dict = doc.to_dict()
            existing_key = as_dict.get(field_name)
            if existing_key is not None:
                existing_unique_doc_ref = self.db.collection(unique_type_name).document(existing_key)
                existing_unique_doc_ref.delete()
        except NotFound:
            # this is fine it just means there is no suxh doc yet
            pass


    def get_unique_instance(self, namespace, type_name, field_name, value):
        unique_type_name = "%s__%s" % (type_name, field_name)
        unique_doc_ref = self.db.collection(unique_type_name).document(value)
        try:
            doc = unique_doc_ref.get()
            as_dict = doc.to_dict()
            wrapper = self._get(as_dict["owner"], as_dict["type_name"])
            return GenericObject._from_doc(self, wrapper.key, wrapper.rev, wrapper.value)

        except NotFound:
            return None


from google.oauth2 import credentials

from firebase_admin.credentials import Base


from google.auth.transport import requests

_request = requests.Request()

_scopes = [
    'https://www.googleapis.com/auth/cloud-platform',
    'https://www.googleapis.com/auth/datastore',
    'https://www.googleapis.com/auth/devstorage.read_write',
    'https://www.googleapis.com/auth/firebase',
    'https://www.googleapis.com/auth/identitytoolkit',
    'https://www.googleapis.com/auth/userinfo.email',
    'https://www.googleapis.com/auth/firebase.database',
]


class CustomToken(Base):
    """A credential initialized from an existing refresh token."""

    _CREDENTIAL_TYPE = 'authorized_user'

    def __init__(self, token, project_id):
        """Initializes a credential from a refresh token JSON file.

        The JSON must consist of client_id, client_secert and refresh_token fields. Refresh
        token files are typically created and managed by the gcloud SDK. To instantiate
        a credential from a refresh token file, either specify the file path or a dict
        representing the parsed contents of the file.

        Args:
          refresh_token: Path to a refresh token file or a dict representing the contents of a
              refresh token file.

        Raises:
          IOError: If the specified file doesn't exist or cannot be read.
          ValueError: If the refresh token configuration is invalid.
        """
        super(CustomToken, self).__init__()

        self._project_id = project_id

        self._g_credential = credentials.Credentials(token=token, scopes=_scopes)

    @property
    def client_id(self):
        return self._g_credential.client_id

    @property
    def client_secret(self):
        return self._g_credential.client_secret

    @property
    def refresh_token(self):
        return self._g_credential.refresh_token

    @property
    def project_id(self):
        return self._project_id


    def get_credential(self):
        """Returns the underlying Google credential.

        Returns:
          google.auth.credentials.Credentials: A Google Auth credential instance."""
        return self._g_credential



import unittest
import json
import time
import os

from google.cloud.firestore_v1beta1 import GeoPoint

import fam
from fam.exceptions import *
from fam.tests.models.test01 import GenericObject, Dog, Cat, Person, NAMESPACE
from fam.tests.models.test04 import House, Fence
from fam.extra_types.lat_long import LatLong

from fam.database import FirestoreWrapper
from fam.mapper import ClassMapper

SECRETS_DIR = os.path.join(os.path.dirname(fam.__file__), "tests", "secrets")


class TestDB(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        creds_path = os.path.join(SECRETS_DIR, "earth-rover-test-d241bce5266d.json")
        mapper = ClassMapper([House, Fence])
        cls.db = FirestoreWrapper(mapper, creds_path, namespace=NAMESPACE)
        cls.clear_db()


    @classmethod
    def clear_db(cls):
        cls.db.delete_all("house")



    def test_geopoint(self):

        loc = LatLong(51.2345, -1.4533)
        house = House.create(self.db, name="my house", location=loc)
        house2 = House.get(self.db, house.key)
        self.assertTrue(type(house2.location) == LatLong)
        self.assertEqual(house2.location.latitude, 51.2345)


    def test_polyline(self):

        loc1 = LatLong(51.2345, -1.4533)
        loc2 = LatLong(51.3345, -1.4533)
        loc3 = LatLong(51.3345, -1.3533)
        loc4 = LatLong(51.2345, -1.3533)

        fence = Fence.create(self.db, name="my house", boundary=[loc1, loc2, loc3, loc4])

        self.assertTrue(type(fence.boundary[2]) == LatLong)



import unittest
import json
import time
import os


import fam
from fam.exceptions import *
from fam.tests.models.test01 import GenericObject, Dog, Cat, Person, JackRussell, Monkey, Monarch, NAMESPACE

from fam.database.mock import MockDatabase
from fam.mapper import ClassMapper



class TestDB(unittest.TestCase):


    def setUp(self):
        mapper = ClassMapper([Dog, Cat, Person, JackRussell, Monkey, Monarch])
        self.db = MockDatabase(mapper)


    def test_app(self):
        self.assertNotEqual(self.db, None)

    def test_make_an_object(self):
        dog = Dog(name="woofer")
        self.assertEqual(dog.namespace, NAMESPACE)
        self.assertEqual(dog.type, "dog")
        self.assertEqual(dog.name, "woofer")
        self.assertEqual(dog.__class__, Dog)

    def test_make_an_object2(self):
        dog = Dog(name="woofer")
        self.assertEqual(dog.talk(), "woof")

    def test_make_a_sub_object(self):
        jack = JackRussell()
        self.assertEqual(jack.talk(), "Yap")
        jack.name = "jack"
        jack.age = 12
        self.db.put(jack)


    def test_make_an_object_saved(self):
        dog = Dog(name="woofer")
        self.db.put(dog)
        self.assertEqual(dog.namespace, NAMESPACE)
        self.assertEqual(dog.type, "dog")
        self.assertEqual(dog.name, "woofer")
        self.assertEqual(dog.__class__, Dog)
        self.assertNotEqual(dog.key, None)
        # self.assertTrue(dog.schema.startswith("http://glowinthedark.co.uk/test/dog"))


    def test_make_an_object_double_dispatch_saved(self):
        dog = Dog(name="woofer")
        self.db.put(dog)
        self.assertEqual(dog.namespace, NAMESPACE)
        self.assertEqual(dog.type, "dog")
        self.assertEqual(dog.name, "woofer")
        self.assertEqual(dog.__class__, Dog)
        self.assertNotEqual(dog.key, None)


    def test_make_an_object_with_additional_properties(self):
        dog = Dog(name="woofer", collar="leather")
        dog.home = "hackney"
        self.db.put(dog)
        self.assertEqual(dog.home, "hackney")

    def test_fail_with_additional_properties(self):

        def wrong_monkey():
            return Monkey(name="bonzo", collar="leather")

        def collar_monkey(monkey):
            monkey.collar = "leather"

        self.assertRaises(FamValidationError, wrong_monkey)
        monkey = Monkey(name="bonzo")
        self.assertRaises(FamValidationError, collar_monkey, monkey)


    def test_get_cat(self):
        paul = Person(name="paul")
        self.db.put(paul)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        self.db.put(cat)
        another = Cat.get(self.db, cat.key)
        self.assertEqual(cat, another)

    def test_get_created_cat(self):
        paul = Person(name="paul")
        self.db.put(paul)
        cat = Cat.create(self.db, name="whiskers", owner_id=paul.key, legs=2)

        another = Cat.get(self.db, cat.key)
        self.assertEqual(cat, another)


    def test_ref_to(self):
        paul = Person(name="paul")
        self.db.put(paul)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        self.db.put(cat)
        self.assertEqual(cat.owner, paul)
        self.assertEqual(cat.owner.name, "paul")


    def test_ref_to_with_object(self):
        paul = Person(name="paul")
        self.db.put(paul)
        cat = Cat(name="whiskers", owner=paul, legs=2)
        self.db.put(cat)
        self.assertEqual(cat.owner, paul)
        self.assertEqual(cat.owner.name, "paul")
        catback = Cat.get(self.db, cat.key)
        self.assertEqual(cat, catback)

    def test_ref_from(self):
        paul = Person(name="paul")
        self.db.put(paul)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        self.db.put(cat)
        cat2 = Cat(name="puss", owner_id=paul.key, legs=2)
        self.db.put(cat2)
        cats = list(paul.cats)
        self.assertEqual(len(cats), 2)
        cats = list(paul.cats)
        self.assertTrue(cats[0] == cat or cats[1] == cat)

    def test_ref_from_multiple_index(self):
        paul = Person(name="paul")
        self.db.put(paul)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        self.db.put(cat)
        dog = Dog(name="fly", owner=paul)
        self.db.put(dog)
        cats = list(paul.cats)
        self.assertEqual(len(cats), 1)
        dogs = list(paul.dogs)
        self.assertEqual(len(dogs), 1)
        animals = list(paul.animals)
        self.assertEqual(len(animals), 2)


    def test_refs_with_inheritance(self):
        paul = Person(name="paul")
        self.db.put(paul)
        jack = JackRussell()
        jack.owner_id = paul.key
        jack.name = "jack"
        jack.save(self.db)
        self.assertEqual(list(paul.dogs)[0], jack)


    def test_refs_with_other_inheritance(self):
        paul = Monarch(name="paul")
        self.db.put(paul)
        jack = Dog()
        jack.owner = paul
        jack.name = "jack"
        self.db.put(jack)
        dogs = list(paul.dogs)
        self.assertEqual(dogs[0], jack)

    def test_delete_cat_dd(self):
        paul = Person(name="paul")
        self.db.put(paul)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        self.db.put(cat)
        key = cat.key
        another = Cat.get(self.db, cat.key)
        self.assertEqual(cat, another)
        self.db.delete(cat)
        revivedcat = self.db.get(key)
        self.assertEqual(revivedcat, None)


    def test_delete_dog_to_refs(self):
        paul = Person(name="paul")
        self.db.put(paul)
        dog = Dog(name="rufus", owner_id=paul.key)
        self.db.put(dog)
        key = dog.key
        dog2 = Dog(name="fly", owner_id=paul.key)
        self.db.put(dog2)
        self.assertTrue(dog2.owner is not None)
        key = paul.key
        self.db.delete(dog)
        revivedpaul = self.db.get(key)
        self.assertTrue(revivedpaul is None)
        refresheddog2 = Dog.get(self.db, dog2.key)
        self.assertTrue(refresheddog2.owner is None)


    def test_delete_cat_refs(self):
        paul = Person(name="paul")
        self.db.put(paul)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        self.assertTrue("dogs" not in cat.__class__.fields.items())
        self.db.put(cat)
        key = cat.key
        cat2 = Cat(name="puss", owner_id=paul.key, legs=2)
        self.assertTrue("dogs" not in cat2.__class__.fields.items())
        self.db.put(cat2)
        revivedcat1 = Cat.get(self.db, key)
        self.assertTrue(revivedcat1 is not None)
        self.db.delete(paul)
        revivedcat2 = self.db.get(key)

        self.assertTrue(revivedcat2 is None)


    def test_delete_dog_refs(self):
        paul = Person(name="paul")
        self.db.put(paul)
        dog = Dog(name="rufus", owner_id=paul.key)
        self.db.put(dog)
        key = dog.key
        dog2 = Dog(name="fly", owner_id=paul.key)
        self.db.put(dog2)
        reviveddog1 = Dog.get(self.db, key)
        self.assertTrue(reviveddog1 is not None)
        self.db.delete(paul)
        reviveddog2 = Dog.get(self.db, key)
        self.assertTrue(reviveddog2 is not None)


    def test_update_cat(self):
        paul = Person(name="paul")
        self.db.put(paul)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        self.db.put(cat)
        cat.name = "blackie"
        self.db.put(cat)

        self.assertEqual(cat.name, "blackie")
        self.assertEqual(cat._properties["name"], "blackie")
        self.assertFalse("name" in cat.__dict__.keys())

        got = Cat.get(self.db, cat.key)

        self.assertEqual(got.name, "blackie")


    def test_update_cat_fails(self):
        paul = Person(name="paul")
        self.db.put(paul)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        self.db.put(cat)
        cat.colour = "black"
        self.db.put(cat)
        def change_colour():
            cat.colour = "white"
        self.assertRaises(FamImmutableError, change_colour)



    def setcatfood(self):
        self.cat.food = "biscuits"


    def test_update_catfood(self):
        paul = Person(name="paul")
        self.db.put(paul)
        cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
        self.db.put(cat)
        self.assertRaises(Exception, self.setcatfood)

    def test_update_dogfood(self):
        dog = Dog(name="fly")
        dog.food = "biscuits"
        self.db.put(dog)


    def test_update_works_without_rev(self):
        monkey = Monkey(name="fly")
        monkey.save(self.db)
        monkey.rev = None
        monkey.name = "jess"
        self.db.put(monkey)


    def test_updates_from_dict(self):

        paul = Person.create(self.db, name="paul")
        dog1 = Dog.create(self.db, name="rufus", owner_id=paul.key, kennel_club_membership="123456")

        attrs = {
            "name":"joe",
            "kennel_club_membership": "9876543"
        }

        dog1.update(attrs)
        dog2 = Dog.get(self.db, dog1.key)
        self.assertTrue(dog2.kennel_club_membership == "9876543")



    # def test_uniqueness(self):
    #     self.clear_db()
    #     paul = Person(name="paul")
    #     self.db.put(paul)
    #     dog1 = Dog.create(self.db, name="rufus", owner_id=paul.key, kennel_club_membership="123456")
    #     dog2 = Dog.create(self.db, name="fly", owner_id=paul.key)
    #
    #     # raises if setting one value
    #     self.assertRaises(FamUniqueError, dog2.update, {"kennel_club_membership": "123456"})
    #     self.assertIsNone(dog2.kennel_club_membership)
    #
    #     # raises if creating a new one
    #     self.assertRaises(FamUniqueError, Dog.create, self.db, name="steve", owner_id=paul.key, kennel_club_membership="123456")
    #
    # def test_uniqueness_delete(self):
    #     self.clear_db()
    #     paul = Person(name="paul")
    #     self.db.put(paul)
    #     dog1 = Dog.create(self.db, name="rufus", owner_id=paul.key, kennel_club_membership="123456")
    #
    #     dog1.delete(self.db)
    #     dog2 = Dog.create(self.db, name="another", owner_id=paul.key, kennel_club_membership="123456")
    #
    #
    # def test_get_unique(self):
    #     self.clear_db()
    #     paul = Person(name="paul")
    #     self.db.put(paul)
    #     dog1 = Dog.create(self.db, name="rufus", owner_id=paul.key, kennel_club_membership="123456")
    #
    #     dog2 = Dog.get_unique_instance(self.db, "kennel_club_membership", "123456")
    #     self.assertIsNotNone(dog2)
    #     self.assertTrue(dog2.kennel_club_membership == "123456")
